#!/usr/bin/env python
"""Prepares an observation for color co-registration and processing.

Prepares images for coregistration. It takes the HiStitch CCD product balance
cubes and scales the BG and IR CCDs to match the corresponding RED CCDs.
If it can, the program deals with these two sets:

- RED4 - BG12 - IR10
- RED5 - BG13 - IR11

HiColorInit must happen after HiccdStitch.


Data Flow
---------
Input Products:

- ``balance.cub`` files for the three or six of the CCDs lsited above, which
    are the result of HiccdStitch.

Output Products:

- A ``.precolor.cub`` file for each of the BG and IR input balance.cub files.

"""

# Copyright 2007-2020, Arizona Board of Regents on behalf of the Lunar and
# Planetary Laboratory at the University of Arizona.
#   - Orignal Perl program.
#
# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#   - Elements of this Python program are are based on the original Perl
#     but the logic here is rewritten from scratch to emulate functionality.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This program is based on HiColor version 5.4.3 (2020/04/28)
# and on the Perl HiColorInit program ($Revision: 1.39 $
#                                      $Date: 2020/04/28 16:56:16 $)
# by Guy McArthur as an employee of the University of Arizona.

import argparse
import logging
import sys
from datetime import datetime
from pathlib import Path

import kalasiris as isis
import hiproc.hirise as hirise
import hiproc.util as util

logger = logging.getLogger(__name__)


CCD_Corresponence = {
    "IR10": "RED4",
    "IR11": "RED5",
    "BG12": "RED4",
    "BG13": "RED5",
}


class HiColorCube(hirise.CCDID):
    """A class for HiRISE CCD IDs with additional capabilities for HiColor."""

    def __init__(self, pathlike):

        self.path = Path(pathlike)
        super().__init__(hirise.get_CCDID_fromfile(self.path))
        self.bin = int(isis.getkey_k(self.path, "Instrument", "Summing"))
        self.tdi = int(isis.getkey_k(self.path, "Instrument", "TDI"))
        self.lines = int(isis.getkey_k(self.path, "Dimensions", "Lines"))
        self.samps = int(isis.getkey_k(self.path, "Dimensions", "Samples"))

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.path}')"


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__,
        parents=[util.parent_parser()],
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "-o", "--output_suffix",
        required=False,
        default=".precolor.cub",
        help="The input color (BG & IR) color cubes will have their final "
             "suffix removed, and this suffix added. Default: %(default)s"
    )
    parser.add_argument(
        "cubes",
        metavar="balance.cub-files",
        nargs="+",
        help="Either one or both sets of RED/IR/BG .balance.cub files. "
             "However, that's tedious to type, so you could just type in "
             "*.balance.cub here, and the program will sort out what it needs."
    )
    return parser


def main():
    args = arg_parser().parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    if not args.output_suffix.startswith("."):
        logger.critical(
            "--output_suffix must start with a period, and it "
            f"does not: {args.output_suffix}"
        )
        sys.exit()

    with util.main_exceptions(args.verbose):
        HiColorInit(args.cubes, args.output_suffix, keep=args.keep)

    return


def HiColorInit(cube_paths: list, output_suffix: str, keep=False):

    logger.info(f"HiColorInit start: {cube_paths}")

    cubes = list(map(HiColorCube, cube_paths))
    (red4, red5, ir10, ir11, bg12, bg13) = separate_ccds(cubes)

    if red4 is None and red5 is None:
        raise ValueError("Neither RED4 nor RED5 were provided.")

    if red4 is not None and red5 is not None:
        if red4.lines != red5.lines or red4.samps != red5.samps:
            raise ValueError("RED4 dimensions not equal to RED5 dimensions!")

    if red4 is not None:
        process_set(red4, ir10, bg12, output_suffix, keep=keep)
    if red5 is not None:
        process_set(red5, ir11, bg13, output_suffix, keep=keep)

    logger.info("HiColorInit done.")
    return


def separate_ccds(cubes: list) -> tuple:
    """Return a tuple of six values, either HiColorCubes, or None."""
    red4 = red5 = ir10 = ir11 = bg12 = bg13 = None
    for c in cubes:
        if c.ccdnumber == "4":
            red4 = c
        elif c.ccdnumber == "5":
            red5 = c
        elif c.ccdnumber == "10":
            ir10 = c
        elif c.ccdnumber == "11":
            ir11 = c
        elif c.ccdnumber == "12":
            bg12 = c
        elif c.ccdnumber == "13":
            bg13 = c
        else:
            # Not one of the six that we care about,
            # so we'll silently ignore it.
            pass
    return red4, red5, ir10, ir11, bg12, bg13


def process_set(
    red: HiColorCube,
    ir: HiColorCube,
    bg: HiColorCube,
    outsuffix: str,
    keep=False,
):
    """Do all of the scaling and cropping for a RED/IR/BG set."""

    if (
        len(
            set(
                map(
                    lambda c: c.get_obsid(),
                    filter(lambda x: x is not None, [red, ir, bg]),
                )
            )
        )
        != 1
    ):
        raise ValueError(
            "These cube files don't all have the same "
            f"Observation ID: {red}, {ir}, {bg}"
        )

    temp_token = datetime.now().strftime("HiColorInit-%y%m%d%H%M%S")

    # These two values are calculated but only written to a PVL file,
    # which I think we can skip.
    # # in bin1 there are 48 pixels of overlap
    # total_width = 2 * red.samps - (48 / red.bin)
    #
    # # in bin1, the right half starts at pixel 2001
    # image_midpoint = 2000 / red.bin + 1

    for c in filter(lambda x: x is not None, [ir, bg]):
        # Calculate delta offset in lines between red and color ccd
        offset = int((200 * (c.bin - red.bin) + c.tdi - red.tdi) / red.bin)

        bin_ratio = c.bin / red.bin
        # tdi_ratio = c.tdi / red.tdi
        mag_ratio = bin_ratio / 1.0006
        # ratio of color to red for enlargement, correction of optical
        # distortion from OPTICAL_ENLARGEMENT_RATIO constant in original
        # HiColor.pm

        # Rescale the color by the bin ratio and mag ratio, to match the red.
        # These will be the BG and IR "pre-color" cubes.
        rescaled = c.path.with_suffix(f".{temp_token}.rescaled" + outsuffix)

        if mag_ratio < 1:
            s = 1 / mag_ratio
            isis.reduce(
                c.path,
                to=rescaled,
                sscale=s,
                lscale=bin_ratio,
                validper=1,
                algorithm="nearest",
                vper_replace="nearest",
            )
        else:
            isis.enlarge(
                c.path,
                to=rescaled,
                sscale=mag_ratio,
                lscale=bin_ratio,
                interp="bilinear",
            )

        # The original Perl had an additional step to divide c.bin by the
        # bin_ratio, and provide that to value= below, but that's
        # mathematically just red.bin, so we'll skip a calculation:
        isis.editlab(
            rescaled,
            options="modkey",
            grpname="Instrument",
            keyword="Summing",
            value=red.bin,
        )

        # trim by placing in a new image
        isis.handmos(
            rescaled,
            mosaic=c.path.with_suffix(outsuffix),
            create="Y",
            nlines=red.lines,
            nsamp=red.samps,
            nband=1,
            outline=offset,
            outsamp=1,
            outband=1,
        )

        if not keep:
            rescaled.unlink()

        logger.info(f"Created {c.path.with_suffix(outsuffix)}")
    return



#!/usr/bin/env python
"""
Perform the jitter derivation for a HiRISE observation.

Outputs are the jitter in the x (sample) and y (line) directions in
pixels, and the time (ephemeris time at that translation), average error
between the derived jitter function and the original data, linerate
(line time read from flat files), and TDI. These outputs are
written to a text file. Another output is the pixel smear, also
written to a file.

The C++ version of this runs ~5x faster, FYI.

WARNING: This program and its modules may not be fully functional and
may return poor results.
"""

# Copyright (C) 2013-2020 Arizona Board of Regents on behalf of the Lunar and
# Planetary Laboratory at the University of Arizona.
#   - Original MatLab program written by Aaron Boyd and Sarah Mattson for
#     HiROC as part of the process to describe and correct geometric
#     distortions caused by jitter in HiRISE images.
#     Original version written approximately 6/2008.
#
#  Copyright (c) 2012, United States Government as represented by the
#  Administrator of the National Aeronautics and Space Administration. All
#  rights reserved.
#   - C++ version written by Oleg Alexandrov based on the above MatLab
#     program, resolveJitter4HiJACK.m version 1.4
#
# Copyright 2020-2021, Ross A. Beyer (rbeyer@seti.org)
#   - Elements of this Python program are are based on the C++ version but
#     the logic here is rewritten from scratch to emulate functionality.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# The only write-up that I could find for the pre-cursor was:
#   S. Mattson, A. Boyd, R. L. Kirk, D. A. Cook, and E. Howington-Kraus,
#   HiJACK: Correcting spacecraft jitter in HiRISE images of Mars,
#   European Planetary Science Congress 2009, #604
#   https://ui.adsabs.harvard.edu/abs/2009epsc.conf..604M}
# However, that's really only a high-level description.

import argparse
import csv
import itertools
import logging
import math
import os
import pkg_resources
import sys
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import medfilt
from scipy.interpolate import PchipInterpolator
from scipy.optimize import minimize_scalar

import pvl

import hiproc.hirise as hirise
import hiproc.util as util
from hiproc.FlatFile import FlatFile

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-c",
        "--conf",
        type=argparse.FileType('r'),
        default=pkg_resources.resource_stream(
            __name__,
            'data/ResolveJitter.conf'
        ),
        help="Path to a ResolveJitter.conf file, only needed if "
        "--lineinterval isn't given.",
    )
    parser.add_argument(
        "--csv",
        action="store_true",
        help="This program writes out a fixed-width data file with extra "
        "information for plotting.  This also writes out a "
        "comma-separated version.",
    )
    parser.add_argument(
        "--lineinterval",
        type=float,
        help="The number of lines to use to set the number of Fourier "
        "transform intervals, defaults to Control_Lines in the "
        "ResolveJitter.conf file.",
    )
    parser.add_argument(
        "--outdir",
        type=Path,
        help="Output directory.  Defaults to the directory of the first "
        "input file.",
    )
    parser.add_argument(
        "--outprefix",
        help="Prefix string for output files.  If not given, will default "
        "to the Observation ID of the images.",
    )
    parser.add_argument(
        "-p", "--plot", action="store_true", help="Displays interactive plot.",
    )
    parser.add_argument(
        "--saveplot",
        nargs="?",
        default=False,
        const=True,
        help="Saves plot to a default filename in the output directory. "
        "If a filename is provided it will be used to save the plot.",
    )
    parser.add_argument(
        "--whichmatch1",
        action="store_false",
        dest="which1",
        help="If specified, the sense of the offsets for the first "
        "file will be relative to the MATCH cube, rather than the FROM "
        "cube.",
    )
    parser.add_argument(
        "--whichmatch2",
        action="store_false",
        dest="which2",
        help="If specified, the sense of the offsets for the second "
        "file will be relative to the MATCH cube, rather than the FROM "
        "cube.",
    )
    parser.add_argument(
        "--whichmatch3",
        action="store_false",
        dest="which3",
        help="If specified, the sense of the offsets for the third "
        "file will be relative to the MATCH cube, rather than the FROM "
        "cube.",
    )
    # parser.add_argument(
    #     "--optimize",
    #     action="store_true",
    #     help="Will run experimental optimizer."
    # )
    parser.add_argument(
        "files",
        nargs="*",
        help="Three flat.txt files that are the output of ISIS hijitreg.",
    )
    return parser


def main():
    parser = arg_parser()
    args = parser.parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    if len(args.files) == 3:
        # With just three arguments, these are the expected flat files.
        fp1, fp2, fp3 = map(Path, args.files)
        which1 = args.which1
        which2 = args.which2
        which3 = args.which3
    elif len(args.files) == 9:
        # This is the old-style positional calling
        oldparser = argparse.ArgumentParser()
        oldparser.add_argument("outdir", type=Path)
        oldparser.add_argument("outprefix", type=str)
        oldparser.add_argument("lineinterval", type=float)
        oldparser.add_argument("file_path1", type=Path)
        oldparser.add_argument("which1", type=int, choices=[-1, 1])
        oldparser.add_argument("file_path2", type=Path)
        oldparser.add_argument("which2", type=int, choices=[-1, 1])
        oldparser.add_argument("file_path3", type=Path)
        oldparser.add_argument("which3", type=int, choices=[-1, 1])
        args = oldparser.parse_args(args.files, namespace=args)
        fp1 = args.file_path1
        fp2 = args.file_path2
        fp3 = args.file_path3
        which1 = True if args.which1 != 1 else False
        which2 = True if args.which2 != 1 else False
        which3 = True if args.which2 != 1 else False
    else:
        parser.error("Only takes 3 or 9 positional arguments.")

    if args.lineinterval is None:
        args.lineinterval = pvl.load(args.conf)["AutoRegistration"][
            "ControlNet"
        ]["Control_Lines"]
    elif args.lineinterval <= 0:
        raise ValueError("--lineinterval must be positive.")

    if args.outdir is None:
        outdir = fp1.parent
    else:
        outdir = args.outdir
        fp1 = set_file_path(outdir, fp1)
        fp2 = set_file_path(outdir, fp2)
        fp3 = set_file_path(outdir, fp3)

    with util.main_exceptions(args.verbose):
        start(
            fp1,
            which1,
            fp2,
            which2,
            fp3,
            which3,
            line_interval=args.lineinterval,
            outdir=outdir,
            outprefix=args.outprefix,
            plotshow=args.plot,
            plotsave=args.saveplot,
            writecsv=args.csv,
            opt=args.optimize
        )
    return


def start(
    file_path1: Path,
    whichfrom1: bool,
    file_path2: Path,
    whichfrom2: bool,
    file_path3: Path,
    whichfrom3: bool,
    line_interval: float,
    outdir: Path,
    outprefix=None,
    plotshow=False,
    plotsave=False,
    writecsv=False,
    opt=False
):
    oid1 = hirise.get_ObsID_fromfile(file_path1)
    oid2 = hirise.get_ObsID_fromfile(file_path2)
    oid3 = hirise.get_ObsID_fromfile(file_path3)

    if oid1 == oid2 == oid3:
        oid = oid1
    else:
        raise ValueError(
            f"The observation IDs from the three file"
            f"paths ({file_path1}, {file_path2}, {file_path3}) do not match."
        )

    (
        nfftime,
        sample,
        line,
        linerate,
        tdi,
        t0,
        t1,
        t2,
        t3,
        offx_filtered,
        offy_filtered,
        xinterp,
        yinterp,
        min_avg_error,
        min_k,
        jittercheckx,
        jitterchecky,
        rh0,
    ) = resolve_jitter(
        file_path1,
        whichfrom1,
        file_path2,
        whichfrom2,
        file_path3,
        whichfrom3,
        line_interval,
        optimize=opt
    )

    # The outputs
    logger.info(f"Average error is {min_avg_error} at min index {min_k}")
    logger.info(f"linerate is {linerate}")
    logger.info(f"TDI = {tdi}")

    # To do: Remove the py suffix from output filenames.

    if outprefix is None:
        outprefix = str(oid)
    else:
        outprefix = str(outprefix)

    # Characterize the smear:
    (
        max_smear_sample,
        max_smear_line,
        max_smear_mag,
        dysdx,
        dyldx,
        xi,
    ) = pixel_smear(nfftime, sample, line, linerate, tdi)

    write_smear_data(
        (outdir / (outprefix + "_smear_py.txt")),
        max_smear_sample,
        max_smear_line,
        max_smear_mag,
        dysdx,
        dyldx,
        xi,
        oid,
    )

    # Make a text file of the jitter data
    jitter_p = outdir / (outprefix + "_jitter_py.txt")
    jitter_text = [
        f"""# Using image {oid} the jitter was found with an
# Average Error of {min_avg_error}
# Maximum Cross-track pixel smear {max_smear_sample}
# Maximum Down-track pixel smear {max_smear_line}
# Maximum Pixel Smear Magnitude {max_smear_mag}
#
# Sample                 Line                   ET"""
    ]
    for s, l, e in zip(sample, line, nfftime):
        jitter_text.append(f"""{s}     {l}     {e}""")

    logger.info(f"Writing: {jitter_p}")
    jitter_p.write_text("\n".join(jitter_text))

    # Create data for plotting
    data_p = outdir / (outprefix + "_jitter_plot_py.txt")
    t_shift = [t1 - t0, t2 - t0, t3 - t0]
    jittercheckx_shift = [
        jittercheckx[0] + rh0["x1"],
        jittercheckx[1] + rh0["x2"],
        jittercheckx[2] + rh0["x3"],
    ]
    jitterchecky_shift = [
        jitterchecky[0] + rh0["y1"],
        jitterchecky[1] + rh0["y2"],
        jitterchecky[2] + rh0["y3"],
    ]

    # Note the comment before the first label
    # This ordering and naming is historic to the original file output.
    string_labels = [
        "# ET_shift",
        "Sample",
        "Line",
        "t1_shift",
        "offx1",
        "xinterp1",
        "jittercheckx1_shift",
        "offy1",
        "yinterp1",
        "jitterchecky1_shift",
        "t2_shift",
        "offx2",
        "xinterp2",
        "jittercheckx2_shift",
        "offy2",
        "yinterp2",
        "jitterchecky2_shift",
        "t3_shift",
        "offx3",
        "xinterp3",
        "jittercheckx3_shift",
        "offy3",
        "yinterp3",
        "jitterchecky3_shift",
    ]
    data_to_plot = (
        string_labels,
        nfftime - t0,
        sample,
        line,
        t_shift[0],
        offx_filtered[0],
        xinterp[0],
        jittercheckx_shift[0],
        offy_filtered[0],
        yinterp[0],
        jitterchecky_shift[0],
        t_shift[1],
        offx_filtered[1],
        xinterp[1],
        jittercheckx_shift[1],
        offy_filtered[1],
        yinterp[1],
        jitterchecky_shift[1],
        t_shift[2],
        offx_filtered[2],
        xinterp[2],
        jittercheckx_shift[2],
        offy_filtered[2],
        yinterp[2],
        jitterchecky_shift[2],
    )

    write_data_for_plotting(data_p, *data_to_plot)

    if writecsv:
        write_csv(outdir / (outprefix + "_jitter_plot_py.csv"), *data_to_plot)

    gnuplot_p = outdir / (outprefix + "_jitter_plot_py.plt")
    img_file_name = outdir / (outprefix + "_jitter_plot_py.png")
    write_gnuplot_file(
        gnuplot_p, data_p, img_file_name, file_path1, file_path2, file_path3
    )

    if plotsave:
        try:
            plotsave = Path(plotsave)
        except TypeError:
            plotsave = outdir / (outprefix + "_jitter_plot_py.pdf")

    if plotshow or plotsave:
        plot(
            t_shift,
            offx_filtered,
            offy_filtered,
            xinterp,
            yinterp,
            jittercheckx_shift,
            jitterchecky_shift,
            [file_path1.stem, file_path2.stem, file_path3.stem],
            nfftime - t0,
            sample,
            line,
            show=plotshow,
            save=plotsave,
        )

    return


def resolve_jitter(
    file_path1: Path,
    whichfrom1: bool,
    file_path2: Path,
    whichfrom2: bool,
    file_path3: Path,
    whichfrom3: bool,
    line_interval: float,
    window_size=11,
    window_width=2,
    optimize=False
):
    """
    Returns a large tuple of information that is the result of solving for
    the jitter based on the three input files.

    The first file path sets some variables for all of the runs.  The
    whichfrom booleans determine determines the sense of the offsets.
    A value of True makes the offsets relative to the From cube specified
    in the flat.tab, and False makes the offsets relative to the Match cube.

    :param file_path1: File Path for first flat.tab file.
    :param whichfrom1: Offsets relative to FROM for first flat.tab file.
    :param file_path2: File Path for second flat.tab file.
    :param whichfrom2: Offsets relative to FROM for second flat.tab file.
    :param file_path3: File Path for third flat.tab file.
    :param whichfrom3: Offsets relative to FROM for third flat.tab file.
    :param line_interval: The number of lines to use to set the number of
        Fourier transform intervals.
    :param window_size: The kernel size for median filtering the offsets,
        should be an odd integer.
    :param window_width: Sets the boundaries above and below the filtered
        average beyond which to exclude outliers.

    :return: A gnarly 18-tuple

    The values of the 18-tuple are:

    0
        Time values starting at *t0* (numpy array).
    1
        Sample direction jitter for each time (numpy array).
    2
        Line direction jitter for each time (numpy array).
    3
        The LineRate from the FROM and MATCH files (float).
    4
        The TDI for the FROM and MATCH files (int).
    5
        Zero time at which the Fourier Transform steps start (float).
    6
        Unique time values from file one (numpy array).
    7
        Unique time values from file two (numpy array).
    8
        Unique time values from file three (numpy array).
    9
        List of three numpy arrays representing the Gaussian filtered
        offsets in the sample direction from the three input files.
    10
        List of three numpy arrays representing the Gaussian filtered
        offsets in the line direction from the three input files.
    11
        List of three numpy arrays representing the interpolation of
        sample offsets for each input file at each time.
    12
        List of three numpy arrays representing the interpolation of
        line offsets for each input file at each time.
    13
        The average error of the jitter solution with respect to the
        measured offsets.
    14
        min_k
    15
        [min_jitter_check_x1, min_jitter_check_x2, min_jitter_check_x3],
    16
        [min_jitter_check_y1, min_jitter_check_y2, min_jitter_check_y3],
    17
        The first value of the sample and line from each file (dict).

    """
    t1, offx1, offy1, lines1, dt1, tdi1, linerate1 = parse_file(
        file_path1, window_size, window_width, whichfrom1
    )
    t2, offx2, offy2, lines2, dt2, tdi2, linerate2 = parse_file(
        file_path2, window_size, window_width, whichfrom2
    )
    t3, offx3, offy3, lines3, dt3, tdi3, linerate3 = parse_file(
        file_path3, window_size, window_width, whichfrom3
    )

    # nfft is the number of "steps" that we will be using for our
    # fourier transforms.  These steps represent a certain number of
    # image lines, but also represent the number of "time steps" that
    # we will use.
    if lines1 == lines2 == lines3:
        nfft = upper_power_of_two(lines1 / line_interval)
    else:
        raise ValueError(
            "The number of lines in the three images is not identical."
        )

    if tdi1 == tdi2 == tdi3:
        tdi = tdi3
    else:
        raise ValueError("The values of tdi are not identical.")

    if linerate1 == linerate2 == linerate3:
        linerate = linerate3
    else:
        raise ValueError("The values of linerate are not identical.")

    offx_filtered = list(
        map(
            filter_data,
            itertools.repeat(nfft),
            itertools.repeat(2 / nfft),
            (offx1, offx2, offx3),
        )
    )
    offy_filtered = list(
        map(
            filter_data,
            itertools.repeat(nfft),
            itertools.repeat(2 / nfft),
            (offy1, offy2, offy3),
        )
    )

    # The values in tt are the fractional points in [0:1]
    # that correspond to the nfft number.
    tt = np.linspace(0, 1, nfft, endpoint=False)

    # The first file to be parsed sets nfftime, t0, and duration
    nfftime = np.linspace(t1[0], t1[-1], nfft)
    # et_shift = et - t1[0]
    t0 = t1[0]
    duration = t1[-1] - t0

    xinterp = [None] * 3
    yinterp = [None] * 3
    xinterp[0], yinterp[0], x1, y1, ddt1, overxx1, overyy1 = create_matrices(
        t1,
        offx_filtered[0],
        offy_filtered[0],
        dt1,
        duration,
        t0,
        nfft,
        nfftime,
        tt,
    )
    xinterp[1], yinterp[1], x2, y2, ddt2, overxx2, overyy2 = create_matrices(
        t2,
        offx_filtered[1],
        offy_filtered[1],
        dt2,
        duration,
        t0,
        nfft,
        nfftime,
        tt,
    )
    xinterp[2], yinterp[2], x3, y3, ddt3, overxx3, overyy3 = create_matrices(
        t3,
        offx_filtered[2],
        offy_filtered[2],
        dt3,
        duration,
        t0,
        nfft,
        nfftime,
        tt,
    )

    logger.info("Searching for correct phasetol")
    # For the test data, the following while loop will *always* run the full
    # number of repetitions.  If it were guaranteed that there was only
    # one minima in the changing value of *error*, then we could exit
    # this loop early, once the error starts going back up, or adopt
    # strategies to divide and conquer the phasetol parameter space.
    # A better understanding of the error behavior could allow us to speed
    # these loops up.
    #
    # Tests with scipy.optimize.minimize_scalar() indicate that this
    # data is not convex, and local minima confuse the minimizers.

    # int      k           = 0;
    # double   minAvgError = numeric_limits<double>::max();
    # int      minK        = 0;
    # double   error       = errorTol;
    # ArrayXd  minJitterX, minJitterY;
    # MatrixXd overxxx1, overyyy1, overxxx2, overyyy2, overxxx3, overyyy3;

    # Tolerance for the error
    error_tol = 0.0000000001
    repetitions = 50

    # Tolerance coefficient for the phase difference
    tolcoef = 0.01

    k = 0
    error = error_tol
    min_avg_error = sys.float_info.max
    # min_k = 0

    # No need to calculate these during every loop:
    rh0 = dict(
        x1=np.real(x1[0]) / 2.0,
        x2=np.real(x2[0]) / 2.0,
        x3=np.real(x3[0]) / 2.0,
        y1=np.real(y1[0]) / 2.0,
        y2=np.real(y2[0]) / 2.0,
        y3=np.real(y3[0]) / 2.0,
    )

    if optimize:
        # the scipy.optimize.minimize_scaler() methods got distracted by
        # a local minima
        opt_res = minimize_scalar(
            jitter_error,
            args=(
                tolcoef,
                tt,
                duration,
                rh0,
                (dt1, dt2, dt3),
                xinterp,
                yinterp,
                (ddt1, ddt2, ddt3),
                (overxx1, overxx2, overxx3),
                (overyy1, overyy2, overyy3)
            ),
            method="Brent",
            bracket=(0, 50),
            # method="bounded",
            # bounds=(0, 50),
            tol=error_tol,
            options=dict(maxiter=repetitions, disp=True)
        )
        logger.info(opt_res)
        min_k = opt_res.x
    else:
        while error >= error_tol and k < repetitions:
            k += 1

            error = jitter_error(
                k,
                tolcoef,
                tt,
                duration,
                rh0,
                (dt1, dt2, dt3),
                xinterp,
                yinterp,
                (ddt1, ddt2, ddt3),
                (overxx1, overxx2, overxx3),
                (overyy1, overyy2, overyy3)
            )

            if error < min_avg_error:
                min_avg_error = error
                min_k = k
        logger.info(f"Minimum Error after phase filtering: {min_avg_error}")

        # end while

    min_jitterx, min_jittery = jitterxy(
        min_k * tolcoef,
        (ddt1, ddt2, ddt3),
        (overxx1, overxx2, overxx3),
        (overyy1, overyy2, overyy3),
    )

    logger.info("Searching for correct filter size.")
    k = 0
    min_k = 0
    error = error_tol
    min_avg_error = sys.float_info.max

    # The jitter in the x (sample) and y (line) directions in pixels.
    # This is the jitter with minimum error, after scanning through all
    # frequencies omega.
    # ArrayXd Sample, Line;
    # ArrayXd  minJitterCheckX1, minJitterCheckX2, minJitterCheckX3;
    # ArrayXd  minJitterCheckY1, minJitterCheckY2, minJitterCheckY3;

    # starting a loop to find the correct filter size
    while error >= error_tol and k < repetitions:

        k += 1

        omega = k - 1
        c = omega / (2.0 * nfft)

        jitterxx = filter_data(nfft, c, min_jitterx)
        jitteryy = filter_data(nfft, c, min_jittery)

        jitterxx = jitterxx - jitterxx[0]
        jitteryy = jitteryy - jitteryy[0]

        # checking
        jittercheckx1 = (
            np.interp(tt + dt1 / duration, tt, jitterxx, left=0, right=0)
            - jitterxx
        )
        jitterchecky1 = (
            np.interp(tt + dt1 / duration, tt, jitteryy, left=0, right=0)
            - jitteryy
        )

        jittercheckx2 = (
            np.interp(tt + dt2 / duration, tt, jitterxx, left=0, right=0)
            - jitterxx
        )
        jitterchecky2 = (
            np.interp(tt + dt2 / duration, tt, jitteryy, left=0, right=0)
            - jitteryy
        )

        jittercheckx3 = (
            np.interp(tt + dt3 / duration, tt, jitterxx, left=0, right=0)
            - jitterxx
        )
        jitterchecky3 = (
            np.interp(tt + dt3 / duration, tt, jitteryy, left=0, right=0)
            - jitteryy
        )

        error_vec = (
            1.0
            / 6.0
            * (
                np.abs(xinterp[0] - (jittercheckx1 + rh0["x1"]))
                + np.abs(xinterp[1] - (jittercheckx2 + rh0["x2"]))
                + np.abs(xinterp[2] - (jittercheckx3 + rh0["x3"]))
                + np.abs(yinterp[0] - (jitterchecky1 + rh0["y1"]))
                + np.abs(yinterp[1] - (jitterchecky2 + rh0["y2"]))
                + np.abs(yinterp[2] - (jitterchecky3 + rh0["y3"]))
            )
        )

        error = error_vec.mean()
        logger.info(f"Erorr for omega {omega}: {error}")

        if error < min_avg_error:
            min_k = k
            min_avg_error = error
            sample = jitterxx
            line = jitteryy
            min_jitter_check_x1 = jittercheckx1
            min_jitter_check_x2 = jittercheckx2
            min_jitter_check_x3 = jittercheckx3
            min_jitter_check_y1 = jitterchecky1
            min_jitter_check_y2 = jitterchecky2
            min_jitter_check_y3 = jitterchecky3
    # end while

    return (
        nfftime,
        sample,
        line,
        linerate,
        tdi,
        t0,
        t1,
        t2,
        t3,
        offx_filtered,
        offy_filtered,
        xinterp,
        yinterp,
        min_avg_error,
        min_k,
        [min_jitter_check_x1, min_jitter_check_x2, min_jitter_check_x3],
        [min_jitter_check_y1, min_jitter_check_y2, min_jitter_check_y3],
        rh0,
    )


def jitterxy(phasetol, ddt, overxx, overyy):
    # it = iter((ddt, overxx, overyy))
    # the_len = len(next(it))
    # if not all(len(l) == the_len for l in it):
    #     raise ValueError('Not all lists have same length!')

    # null the frequencies that cause a problem (really zero them out)
    masked_overxx = list()
    masked_overyy = list()
    for (d, x, y) in zip(ddt, overxx, overyy):
        xxx, yyy = mask_frequencies(phasetol, d, x, y)
        masked_overxx.append(xxx)
        masked_overyy.append(yyy)

    # Adding all frequencies together
    stackedx = np.ma.stack(masked_overxx)
    stackedy = np.ma.stack(masked_overyy)

    overxxx = np.ma.mean(stackedx, axis=0)
    overyyy = np.ma.mean(stackedy, axis=0)

    # take the sum of each row
    overx = np.ma.sum(overxxx, axis=0)
    overy = np.ma.sum(overyyy, axis=0)

    jitterx = overx - overx[0]
    jittery = overy - overy[0]

    return jitterx, jittery


def jitter_error(
    k, tolcoef, tt, duration, rh0, dt, xinterp, yinterp, ddt, overxx, overyy
):
    # setting the phase tolerance
    phasetol = k * tolcoef

    jitterx, jittery = jitterxy(
        phasetol,
        ddt,
        overxx,
        overyy,
    )

    # checking
    jittercheckx1 = (
        np.interp(tt + dt[0] / duration, tt, jitterx, left=0, right=0)
        - jitterx
    )
    jitterchecky1 = (
        np.interp(tt + dt[0] / duration, tt, jittery, left=0, right=0)
        - jittery
    )

    jittercheckx2 = (
        np.interp(tt + dt[1] / duration, tt, jitterx, left=0, right=0)
        - jitterx
    )
    jitterchecky2 = (
        np.interp(tt + dt[1] / duration, tt, jittery, left=0, right=0)
        - jittery
    )

    jittercheckx3 = (
        np.interp(tt + dt[2] / duration, tt, jitterx, left=0, right=0)
        - jitterx
    )
    jitterchecky3 = (
        np.interp(tt + dt[2] / duration, tt, jittery, left=0, right=0)
        - jittery
    )

    error_vec = (
                    np.abs(xinterp[0] - (jittercheckx1 + rh0["x1"]))
                    + np.abs(xinterp[1] - (jittercheckx2 + rh0["x2"]))
                    + np.abs(xinterp[2] - (jittercheckx3 + rh0["x3"]))
                    + np.abs(yinterp[0] - (jitterchecky1 + rh0["y1"]))
                    + np.abs(yinterp[1] - (jitterchecky2 + rh0["y2"]))
                    + np.abs(yinterp[2] - (jitterchecky3 + rh0["y3"]))
                ) / 6.0

    error = error_vec.mean()
    logger.info(f"Error for phasetol {phasetol}: {error}")
    return error


def create_matrices(
    time: np.array,
    offx: np.array,
    offy: np.array,
    dt: float,
    duration: float,
    t0: float,
    nfft: int,
    nfftime: np.array,
    tt: np.array,
):
    """Returns a tuple of numpy arrays.

    :param time: Time values in seconds (numpy array).
    :param offx: Sample offsets (numpy array).
    :param offy: Line offsets (numpy array).
    :param dt: Time difference between the FROM and MATCH times (float).
    :param duration: The total time duration (float).
    :param t0: Zero time to start the Fourier Transform steps (float).
    :param nfft: Number of divisions to use for the Fourier Transform (int).
    :param nfftime: *nfft* time values starting at *t0* (numpy array).
    :param tt: Fractional values in [0:1] that correspond to the nfft number
        (numpy array).
    :return: There are seven elements in the tuple:
        0: Interpolation of *offx* at the values of *nfftime*
        1: Interpolation of *offy* at the values of *nfftime*
        2: Fourier transform of xinterp * (2 / *nfft*)
        3: Fourier transform of yinterp * (2 / *nfft*)
        4: phase difference
        5: overxx: ?
        6: overyy: ?
    """
    t_shift = time - t0

    fx = PchipInterpolator(t_shift, offx, extrapolate=False)
    fy = PchipInterpolator(t_shift, offy, extrapolate=False)
    xinterp = fx(nfftime - t0)
    yinterp = fy(nfftime - t0)

    np.nan_to_num(xinterp, copy=False, nan=np.mean(offx))
    np.nan_to_num(yinterp, copy=False, nan=np.mean(offy))

    # getting the frequencies of the Fourier transform
    freq = np.linspace(0, nfft / 2, int(nfft / 2), endpoint=False)

    # taking the fourier transform of the offsets
    x = 2 * np.fft.fft(xinterp) / nfft
    y = 2 * np.fft.fft(yinterp) / nfft

    # separating sines and cosines
    xa = x[: int(nfft / 2)].real
    xb = -1 * x[: int(nfft / 2)].imag
    ya = y[: int(nfft / 2)].real
    yb = -1 * y[: int(nfft / 2)].imag

    # calculates the phase difference
    twopi = math.pi * 2
    ddt_temp = (dt / duration * twopi) * freq
    ddt = ddt_temp - twopi * np.floor(ddt_temp / twopi)

    # the coeficients for the frequencies
    with np.errstate(divide="ignore"):
        aaax = (
            -0.5
            * (-1 * xa * np.cos(ddt) + np.sin(ddt) * xb - xa)
            / np.sin(ddt)
        )
        aaay = (
            -0.5
            * (-1 * ya * np.cos(ddt) + np.sin(ddt) * yb - ya)
            / np.sin(ddt)
        )

    with np.errstate(invalid="ignore"):
        bbbx = -0.5 * (xb * np.cos(ddt) + np.sin(ddt) * xa + xb) / np.sin(ddt)
        bbby = -0.5 * (yb * np.cos(ddt) + np.sin(ddt) * ya + yb) / np.sin(ddt)

    # create series of sines and cosines
    ft = freq.reshape(-1, 1) * tt.reshape(1, -1) * twopi
    sn = np.sin(ft)
    cn = np.cos(ft)
    aaax_rep = np.repeat(aaax.reshape(-1, 1), tt.size, axis=1)
    bbbx_rep = np.repeat(bbbx.reshape(-1, 1), tt.size, axis=1)
    aaay_rep = np.repeat(aaay.reshape(-1, 1), tt.size, axis=1)
    bbby_rep = np.repeat(bbby.reshape(-1, 1), tt.size, axis=1)
    with np.errstate(invalid="ignore"):
        overxx = aaax_rep * sn + bbbx_rep * cn
        overyy = aaay_rep * sn + bbby_rep * cn

    # Outputs
    # ArrayXd & tt, ArrayXd & ET, ArrayXd & ET_shift,
    # ArrayXd & ddt,
    # ArrayXd & xinterp, ArrayXd & yinterp,
    # ArrayXcd & X, ArrayXcd & Y, MatrixXd & overxx, MatrixXd & overyy
    #
    return xinterp, yinterp, x, y, ddt, overxx, overyy


def upper_power_of_two(value) -> int:
    """Returns the value of 2 raised to some power which is the smallest
     such value that is just >= *value*."""
    result = 1
    while result < value:
        result <<= 1
    return result


def filter_data(nfft: int, c: float, data: np.array) -> np.array:
    """Apply a Gaussian filter to the data in the frequency domain.

    :param nfft: The number of steps in the Fourier Transform (int).
    :param c: ?
    :param data: An array of data to be filtered (numpy array).
    """

    if len(data.shape) > 1:
        raise IndexError("The data array can only be 1D.")

    # Use padding so the data is not distorted.
    front_padding = math.floor(nfft - len(data) / 2)
    back_padding = math.ceil(nfft - len(data) / 2)

    # Apply the padding
    padded = np.concatenate(
        (
            np.array([data[0]] * front_padding),
            data,
            np.array([data[-1]] * back_padding),
        )
    )

    freq = np.fft.fft(padded)

    # The exponential
    exp_vec = np.hstack(
        (np.linspace(0, nfft - 1, nfft), np.linspace(-1 * nfft, -1, nfft))
    )

    exponential = np.exp(-1 * c ** 2 * exp_vec ** 2)

    # The ifft of the product
    filtered = np.fft.ifft(freq * exponential)

    # Remove the padding and take the real part
    return filtered[front_padding : front_padding + len(data)].real


def mask_frequencies(phasetol: float, ddt: np.array, x: np.array, y: np.array):
    """Returns *x* and *y* as numpy masked arrays with 'problematic frequencies'
    masked.

    :param phasetol: Phase values from zero to *phasetol* and 2*pi - phasetol
        will be masked.
    :param ddt: phase difference
    :param x: overxx from create_matrices()
    :param y: overyy from create_matrices()

    It is assumed that *ddt* has the same size as axis 0 of *x* and *y*.
    """
    if x.shape != y.shape:
        raise ValueError(
            f"The shape of x {x.shape} and y {y.shape} must be the same."
        )

    if x.shape[0] != ddt.size:
        raise ValueError(
            f"The size of ddt ({ddt.size}) must be the same as the first axis"
            f"of the x and y arrays {x.shape}"
        )

    # mask the frequencies that cause a problem
    a = np.less(np.abs(ddt), phasetol)
    b = np.greater(np.abs(ddt), (2 * math.pi) - phasetol)
    null_positions = np.logical_or(a, b)

    # We must reshape to a 2D column, and then tile that across, so that each
    # row of the 2D matrix has the same value for all positions.
    null_2d = np.tile(null_positions.reshape(-1, 1), (1, x.shape[1]))

    x_masked = np.ma.array(x, mask=null_2d, fill_value=0)
    y_masked = np.ma.array(y, mask=null_2d, fill_value=0)

    return x_masked, y_masked


def parse_file(
    file_path: os.PathLike, window_size: int, window_width: int, whichfrom=True
):
    """Returns a tuple of information from the Flat file at *file_path*.

    There are seven elements in the tuple:
    0: unique time values (numpy array)
    1: sample offsets for each time (numpy array)
    2: line offsets for each time (numpy array)
    3: number of lines listed for the FROM file (int)
    4: seconds between the times in the FROM and MATCH files (float)
    5: The TDI for the FROM and MATCH files (int)
    6: The LineRate from the FROM and MATCH files (float)

    *window_size* is the kernel size for median filtering the offsets, should
    be an odd integer.

    *window_width* determines the boundaries above and below the filtered
    average beyond which to exclude outliers.

    The optional *whichfrom* parameter determines the sense of the offsets.
    The default value of True makes the offsets relative to the From cube,
    and False makes the offsets relative to the Match cube.
    """
    logger.info(f"Reading: {file_path}")

    flat = FlatFile(file_path)

    if flat["FROM"]["TdiMode"] == flat["MATCH"]["TdiMode"]:
        tdi = int(flat["FROM"]["TdiMode"])
    else:
        raise ValueError(
            f"The TdiMode is different for FROM ({flat['FROM']['TdiMode']}) "
            f"and MATCH ({flat['MATCH']['TdiMode']}) in {file_path}"
        )

    if flat["FROM"]["LineRate"] == flat["MATCH"]["LineRate"]:
        line_rate = float(flat["FROM"]["LineRate"].split()[0])
    else:
        raise ValueError(
            f"The LineRate is different for FROM ({flat['FROM']['LineRate']}) "
            f"and MATCH ({flat['MATCH']['LineRate']}) in {file_path}"
        )

    if whichfrom == 1:
        column = "FromTime"
        which = -1
    else:
        column = "MatchTime"
        which = 1

    # dt = which * (data[0][0] - data[0][3]);
    dt = which * (float(flat[0]["FromTime"]) - float(flat[0]["MatchTime"]))

    time = list()
    offset_x = list()
    offset_y = list()

    for row in flat:
        time.append(float(row[column]))
        offset_x.append(
            which * (float(row["RegSamp"]) - float(row["FromSamp"]))
        )
        offset_y.append(
            which * (float(row["RegLine"]) - float(row["FromLine"]))
        )

    time_arr = np.array(time)
    offx_arr = np.array(offset_x)
    offy_arr = np.array(offset_y)

    magnitude = np.sqrt(offx_arr ** 2 + offy_arr ** 2)
    avemag = medfilt(magnitude, window_size)

    # Throw out the out-of-range values:
    high_window = avemag + window_width
    low_window = avemag - window_width
    good_idxs = np.nonzero(
        np.logical_and(low_window < magnitude, magnitude < high_window)
    )

    # Some flat files have more than one measurement for the same timestamp.
    # For those that do, the multiple x and y offsets for the same
    # timestamp are averaged together.
    #
    # Also, the original MatLab code had no functions, but repeated code,
    # and there was a transcription error on the third pass of this averaging
    # such that the very first row was never included in the averaging.
    # The C++ code allowed for that broken behavior, but we won't here.
    t_arr, unique_idxs = np.unique(time_arr[good_idxs], return_index=True)

    if unique_idxs.size == time_arr[good_idxs].size:
        offx = offx_arr[good_idxs]
        offy = offy_arr[good_idxs]
    else:
        x_means = list()
        for a in np.split(offx_arr[good_idxs], unique_idxs[1:]):
            x_means.append(np.mean(a))
        offx = np.array(x_means)

        y_means = list()
        for a in np.split(offy_arr[good_idxs], unique_idxs[1:]):
            y_means.append(np.mean(a))
        offy = np.array(y_means)

    return t_arr, offx, offy, int(flat["FROM"]["Lines"]), dt, tdi, line_rate


def pixel_smear(
    t: np.array, sample: np.array, line: np.array, linerate: float, tdi: int,
):
    """Returns the smear values from the derived jitter function.

    Pixel smear due to jitter is calculated by interpolating the jitter
    function at intervals equivalent to the linerate.  Then the
    difference is taken over that interval and multiplied by the TDI.
    This provides an estimated minimum for pixel smear. If the motion
    that caused the smear is not captured in the jitter derivation,
    then it cannot be plotted here.

    :param t: Time values (numpy array).
    :param sample: Pixel offsets in the sample direction (numpy array).
    :param line: Pixel offsest in the line direction (numpy array).
    :param linerate: The image line rate (float).
    :param tdi: The image TDI value (int).
    :param path: Optional path to write out smear details to (Path).
    :param image_id: Optional Observation ID (string).
    :return: six-tuple which contains:
        0. sample maximum smear value (float)
        1. line maximum smear value (float)
        2. Maximum smear magnitude (float)
        3. Sample smear values (numpy array)
        4. Line smear values (numpy array)
        5. Ephemeris times at linerate intervals (numpy array)
    """
    # The output here differs from the C++ output in a very, very minor
    # way.  The absolute difference between the line and sample smear
    # is 0.002 pixels.  I could not track this down.  It may be due to
    # precision differences in the double values that C++ uses, but I'm
    # just not sure.

    # The array T has large values. Use a shifted version of it when
    # interpolating to reduce the effect of those values on numerical
    # accuracy.
    shifted_t = t - t[0]

    # xi = T(1):linerate:T(end);
    n = math.floor((shifted_t[-1] - shifted_t[0]) / linerate) + 1
    # This is from the original code, but by definition shifted_t[0] is zero.
    # xi = np.linspace(0, n - 1, n) * linerate + shifted_t[0]
    xi = np.linspace(0, n - 1, n) * linerate

    # Interpolate the jitter function at intervals equivalent to the linerate
    f_samp = PchipInterpolator(shifted_t, sample, extrapolate=False)
    f_line = PchipInterpolator(shifted_t, line, extrapolate=False)
    yis = f_samp(xi)
    yil = f_line(xi)

    np.nan_to_num(yis, copy=False, nan=0)
    np.nan_to_num(yil, copy=False, nan=0)

    # Undo the earlier shift
    xi += t[0]

    # Calculate the rate of change with respect to the linerate
    # in the sample direction
    dysdx = np.diff(yis) * tdi
    # in the line direction
    dyldx = np.diff(yil) * tdi

    # Calculate the magnitude of the smear
    mag_smear = np.sqrt(dysdx ** 2 + dyldx ** 2)

    # Find maxSmearS, the largest element by magnitude in dysdx
    msi = np.argmax(np.abs(dysdx))
    max_smear_s = dysdx[msi]

    # Find maxSmearL, the largest element by magnitude in dyldx
    msi = np.argmax(np.abs(dyldx))
    max_smear_l = dyldx[msi]

    # Find maxSmearMag, the largest element by magnitude in magSmear
    max_smear_mag = np.max(mag_smear)

    # Outputs
    return max_smear_s, max_smear_l, max_smear_mag, dysdx, dyldx, xi


def set_file_path(location: Path, file_path: Path):
    if file_path.parent == location:
        return file_path
    else:
        return location / file_path.name


def plot(
    t,
    x,
    y,
    xinterp,
    yinterp,
    jittercheckx,
    jitterchecky,
    title,
    et,
    sample,
    line,
    show=True,
    save=False,
):
    plt.ioff()
    fig = plt.figure(constrained_layout=True)
    gs = fig.add_gridspec(3, 6)

    fig.suptitle("Resolve Jitter Results")

    ax00 = fig.add_subplot(gs[0, 0:2])
    ax00.set_title(title[0])
    ax00.set_ylabel("Sample Offset")

    ax01 = fig.add_subplot(gs[0, 2:4])
    ax01.set_title(title[1])

    ax02 = fig.add_subplot(gs[0, 4:])
    ax02.set_title(title[2])

    ax10 = fig.add_subplot(gs[1, 0:2])
    ax10.set_ylabel("Line Offset")
    ax10.set_xlabel("Seconds")
    ax11 = fig.add_subplot(gs[1, 2:4])
    ax11.set_xlabel("Seconds")
    ax12 = fig.add_subplot(gs[1, 4:])
    ax12.set_xlabel("Seconds")

    ax20 = fig.add_subplot(gs[2, 0:3])
    ax20.set_title("Cross-Track Jitter")
    ax20.set_ylabel("Sample Offset")
    ax20.set_xlabel("Seconds")

    ax21 = fig.add_subplot(gs[2, 3:])
    ax21.set_title("Down-Track Jitter")
    ax21.set_ylabel("Line Offset")
    ax21.set_xlabel("Seconds")

    ax00.plot(t[0], x[0], "o", c="red")
    ax00.plot(et, xinterp[0], c="green")
    ax00.plot(et, jittercheckx[0], c="yellow")

    ax01.plot(t[1], x[1], "o", c="red")
    ax01.plot(et, xinterp[1], c="green")
    ax01.plot(et, jittercheckx[1], c="yellow")

    ax02.plot(t[2], x[2], "o", c="red")
    ax02.plot(et, xinterp[2], c="green")
    ax02.plot(et, jittercheckx[2], c="yellow")

    ax10.plot(t[0], y[0], "o", c="red")
    ax10.plot(et, yinterp[0], c="green")
    ax10.plot(et, jitterchecky[0], c="yellow")

    ax11.plot(t[1], y[1], "o", c="red")
    ax11.plot(et, yinterp[1], c="green")
    ax11.plot(et, jitterchecky[1], c="yellow")

    ax12.plot(t[2], y[2], "o", c="red")
    ax12.plot(et, yinterp[2], c="green")
    ax12.plot(et, jitterchecky[2], c="yellow")

    ax20.plot(et, sample, c="blue")
    ax21.plot(et, line, c="blue")

    if save:
        logger.info(f"Writing: {save}")
        plt.savefig(save)

    if show:
        plt.show()


def write_csv(path: os.PathLike, labels: list, *cols, fillvalue="nan"):
    """Identical to write_data_for_plotting(), but writes a CSV file
    instead of a fixed-width text file.
    """
    logger.info(f"Writing: {path}")
    with open(path, "w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(labels)
        for zipped in itertools.zip_longest(*cols, fillvalue=fillvalue):
            writer.writerow(zipped)


def write_data_for_plotting(
    path: os.PathLike, labels: list, *cols, fillvalue="nan"
):
    """Given a collection of arrays of data (*cols*), write those arrays as
    fixed-width columns (25 characters wide) in a text file at *path*, prefixed
    by the *labels* in the first row. The columns need not have the same number
    of elements as each other, and any short columns will be filled with
    *fillvalue*.

    This is mostly historic to provide a text file that gnuplot can use.
    """

    if len(labels) != len(cols):
        raise IndexError(
            "There is a different number of column labels than columns."
        )

    logger.info(f"Writing: {path}")
    with open(path, "w") as f:
        f.write("".join(map(lambda s: s.ljust(25), labels)) + "\n")

        # Print the data columns
        for zipped in itertools.zip_longest(*cols, fillvalue=fillvalue):
            f.write(
                "".join(map(lambda z: "{:>25.16}".format(z), zipped)) + "\n"
            )

        f.write("\n")


def write_smear_data(
    path: Path,
    max_smear_s,
    max_smear_l,
    max_smear_mag,
    dysdx,
    dyldx,
    et,
    image_id=None,
):
    # Make a text file of the smear data
    if image_id is None:
        id_str = ""
    else:
        id_str = f" for {image_id}"

    smear_text = [
        f"""\
# Smear values are calculated from the derived jitter function{id_str}.
# Maximum Cross-track pixel smear {max_smear_s}
# Maximum Down-track pixel smear {max_smear_l}
# Maximum Pixel Smear Magnitude {max_smear_mag}
# Sample                 Line                   EphemerisTime"""
    ]

    for ess, ell, exi in zip(dysdx, dyldx, et):
        smear_text.append(f"{ess}     {ell}     {exi}")

    logger.info(f"Writing: {path}")
    path.write_text("\n".join(smear_text))

    return


def write_gnuplot_file(
    gnuplot_path: Path,
    data_path: Path,
    img_path: Path,
    file_path1: Path,
    file_path2: Path,
    file_path3: Path,
):
    """Writes a gnuplot file that will plot the contents of a file
    written by write_data_for_plotting().

    The file will be written to *gnuplot_path*.  *data_path* should be the
    file written by write_data_for_plotting().  The *img_path* is the png
    file that will be created when the file at *gnuplot_path* is run by
    gnuplot.  *file_path1*, *file_path2*, and *file_path3* are just used
    to provide titles to the plots.
    """
    logger.info(f"Writing: {gnuplot_path}")
    gnuplot_path.write_text(
        f"""\
dataFile  = '{data_path}'
imgFile   = '{img_path}'
filePath1 = '{file_path1}'
filePath2 = '{file_path2}'
filePath3 = '{file_path3}'

set terminal png size 1200, 900; set output imgFile
#set terminal pdfcairo;           set output 'fig.pdf'

set multiplot  # get into multiplot mode
set nokey      # no legend
set grid

set datafile missing 'nan'

w3 = 1.0/3.0; # will do 3 columns of plots
set size w3, w3

set title filePath1
set origin 0,    2*w3
plot dataFile using 4:5 with points pointtype 7 pointsize 0.6 lc rgb 'red', \
dataFile using 1:6 with lines lc rgb 'green', \
dataFile using 1:7 with lines lc rgb 'yellow'

set title filePath2
set origin w3,   2*w3
plot dataFile using 11:12 with points pointtype 7 pointsize 0.6 lc rgb 'red', \
dataFile using 1:13 with lines lc rgb 'green', \
dataFile using 1:14 with lines lc rgb 'yellow'

set title filePath3
set origin 2*w3, 2*w3
plot dataFile using 18:19 with points pointtype 7 pointsize 0.6 lc rgb 'red', \
dataFile using 1:20 with lines lc rgb 'green', \
dataFile using 1:21 with lines lc rgb 'yellow'

set title ''
set origin 0,     w3
plot dataFile using 4:8 with points pointtype 7 pointsize 0.6 lc rgb 'red', \
dataFile using 1:9 with lines lc rgb 'green', \
dataFile using 1:10 with lines lc rgb 'yellow'

set title ''
set origin w3,    w3
plot dataFile using 11:15 with points pointtype 7 pointsize 0.6 lc rgb 'red', \
dataFile using 1:16 with lines lc rgb 'green', \
dataFile using 1:17 with lines lc rgb 'yellow'

set title ''
set origin 2*w3,  w3
plot dataFile using 18:22 with points pointtype 7 pointsize 0.6 lc rgb 'red', \
dataFile using 1:23 with lines lc rgb 'green', \
dataFile using 1:24 with lines lc rgb 'yellow'

w2 = 0.5 # 1/2 of the plotting window
set size w2, w3

set title 'Cross-track Jitter'
set origin 0,     0
plot dataFile using 1:2 with lines lc rgb 'blue'

set title 'Down-track Jitter'
set origin w2,    0
plot dataFile using 1:3 with lines lc rgb 'blue'

unset multiplot # exit multiplot mode
"""
    )
    return



#!/usr/bin/env python
"""This module contains PyRISE utility functions."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import collections
import logging
import os
import subprocess
import sys
import textwrap
from pathlib import Path

import hiproc
import hiproc.hirise as hirise


# These are primarily for use by HiCal and lisfix, primarily meant to be
# used by pause_slicer()
# 1st pixel = index 1 for these
ch_pause = (
    (252, 515, 778),  # Channel 0 pause point sample locations
    (247, 510, 773),  # Channel 1 pause point sample locations
)
ch_width = (
    (17, 17, 17),  # Number of pixels to cut from pause point
    (-17, -17, -17),
)


class main_exceptions:
    """This is a context manager that does the "standard handling"
    of exceptions at the top level in the main() functions in all of
    the programs in this library.  This is just a way to keep that code
    in one place rather than having to repeat the identical try-except blocks
    in each program.

    Using this context manager will cause information about exceptions to
    be printed to sys.stderr (or an alternate file-like of your choosing)
    and may cause sys.exit() to be called.

    Use it like this::

     with util.main_exceptions(args.verbose):
         isis.cubeit(fromlist=f, to='stacked.cub')

    """

    def __init__(self, verbose=0, file=sys.stderr, logger=None):
        self.verbose = verbose
        self.file = file
        self.logger = logger

    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            if issubclass(exc_type, subprocess.CalledProcessError):
                print(
                    textwrap.dedent(
                        f"""\
                        Had a subprocess error:
                        {' '.join(exc_val.cmd)}
                        {exc_val.stdout}
                        {exc_val.stderr}
                        """
                    ),
                    file=self.file
                )
                if self.verbose >= 2:
                    return False
                else:
                    sys.exit(exc_val.returncode)
            if issubclass(exc_type, Warning):
                if self.logger is None:
                    print(exc_val, file=self.file)
                else:
                    self.logger.warning(exc_val)
                return True
            else:
                print("Error:", file=self.file)
                print(exc_val, file=self.file)
                if self.verbose >= 2:
                    return False
                else:
                    sys.exit(1)

        return


def parent_parser() -> argparse.ArgumentParser:
    """Returns a parent parser with common arguments for PyRISE programs."""
    parent = argparse.ArgumentParser(add_help=False)
    parent.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Displays additional information as processing progresses."
    )
    parent.add_argument(
        "-l",
        "--log",
        required=False,
        help=argparse.SUPPRESS,
        # help="The log level to show for this program, can "
        # "be a named log level or a numerical level.",
    )
    parent.add_argument(
        "--logfile",
        required=False,
        help="The log file to write log messages to in addition "
        "to the terminal.",
    )
    parent.add_argument(
        "-k",
        "--keep",
        required=False,
        default=False,
        action="store_true",
        help="Normally, the program will clean up any "
        "intermediary files, but if this option is given, it "
        "won't.",
    )
    parent.add_argument(
        '--version',
        action='version',
        version=f"hiproc version {hiproc.__version__}",
        help="Show hiproc version number."
    )
    return parent


def set_logger(verblvl=None, filename=None, loglvl=0) -> None:
    """Sets the log level and configuration for applications."""
    logger = logging.getLogger(__name__.split(".")[0])
    if loglvl is not None:
        if isinstance(loglvl, int):
            lvl = loglvl
        else:
            lvl = getattr(logging, loglvl.upper(), logging.WARNING)
    else:
        lvl_dict = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}
        if verblvl in lvl_dict:
            lvl = lvl_dict[verblvl]
        else:
            lvl = lvl_dict[max(lvl_dict.keys())]

    logger.setLevel(lvl)

    ch = logging.StreamHandler()
    ch.setLevel(lvl)

    if lvl < 20:  # less than INFO
        formatter = logging.Formatter("%(name)s - %(levelname)s: %(message)s")
        # kalasiris logger
        k_logger = logging.getLogger("kalasiris")
        k_logger.setLevel(lvl)
        k_logger.addHandler(ch)
    else:
        formatter = logging.Formatter("%(levelname)s: %(message)s")

    ch.setFormatter(formatter)
    logger.addHandler(ch)

    if filename is not None:
        fh = logging.FileHandler(filename)
        fh.setLevel(lvl)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
        if lvl < 20:
            k_logger.addHandler(fh)

    return


def path_w_suffix(in_path: str, template_path: os.PathLike) -> Path:
    """If the input starts with a '.' assume it is a suffix and return the
    template with the suffix replaced, otherwise return the input."""
    if in_path.startswith("."):
        return Path(template_path).with_suffix(in_path)
    else:
        return Path(in_path)


def pid_path_w_suffix(in_path: str, template_path: os.PathLike) -> Path:
    """A little extra twist to look for the db file."""
    p = path_w_suffix(in_path, template_path)
    if p.exists():
        return p
    elif in_path.startswith("."):
        pid = hirise.get_ChannelID_fromfile(template_path)
        t_path = Path(template_path)
        if t_path.is_dir():
            d = t_path
        else:
            d = t_path.parent
        pid_path = d / Path(str(pid)).with_suffix(in_path)
        if pid_path.exists():
            return pid_path
        else:
            raise FileNotFoundError(f"Could not find {pid_path}")
    else:
        raise FileNotFoundError(f"Could not find {p}")


def get_path(in_path: os.PathLike, search=None) -> Path:
    """Returns a path that can resovled in the search path
    (or list of paths)."""
    in_p = Path(in_path)
    if in_p.exists():
        return in_p

    search_paths = list()
    if isinstance(search, str):
        search_paths.append(Path(search))
    elif isinstance(search, Path):
        search_paths.append(search)
    elif isinstance(search, collections.abc.Sequence):
        for s in search:
            search_paths.append(Path(s))
    elif search is None:
        raise ValueError("You must provide a path or list of paths to search.")
    else:
        raise TypeError(
            f"Unfortunately, {search} isn't an os.PathLike or a list of them."
        )

    for p in search_paths:
        if p.is_dir():
            out_p = p / in_p
            if out_p.exists():
                return out_p
        else:
            raise NotADirectoryError(f"{p} is not a directory.")
    else:
        raise FileNotFoundError(f"Could not find {in_p} in {search_paths}.")


def conf_check_strings(
    conf_name: str, choices: tuple, conf_value: str
) -> None:
    assert (
        conf_value in choices
    ), f"The {conf_name} parameter can be {choices}, but was {conf_value}"


def conf_check_bool(conf_name: str, conf_value: bool) -> None:
    assert isinstance(conf_value, bool), (
        f"The {conf_name} parameter must be boolean but is {type(conf_value)} "
        f"with value {conf_value}."
    )


def conf_check_count(
    conf_name: str, count: int, what: str, conf_value: list
) -> None:
    assert len(conf_value) == count, (
        f"The {conf_name} parameter must have {count} entries, one for each "
        f"{what}, but it was {conf_value}"
    )


def conf_check_bounds(conf_name: str, bounds: tuple, conf_value: str) -> None:
    assert bounds[0] <= float(conf_value) <= bounds[1], (
        "The {} parameter must be between {} and {} inclusive, "
        "but was {}".format(conf_name, *bounds, conf_value)
    )


def pause_slicer(samp: int, width: int) -> slice:
    """Returns a slice object which satisfies the range of indexes for a pause
    point.

    The incoming numbers for samp are 1-based pixel numbers, so must
    subtract 1 to get a list index.
    The width values are the number of pixels to affect, including the
    pause point pixel.  If positive they start with the pause point pixel
    and count 'up.'  If negative, they start with the pause point pixel
    and count 'down.'
    """
    # We don't need to protect for indices less than zero or greater than the
    # length of the list, because slice objects can take values that would not
    # be valid for item access.
    if width > 0:
        s_start = samp - 1
        s_stop = s_start + width
    else:
        s_start = samp + width
        s_stop = samp
    return slice(s_start, s_stop)



#!/usr/bin/env python
"""Takes the <ObsID>_smear_[cpp|py].txt files and plots two at the
same time for comparison.
"""

import argparse

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np


def read_data(path: Path) -> dict:
    s = list()
    el = list()
    e = list()
    with open(path) as f:
        # for i in range(10000):
        #    line = f.readline()
        for line in f:
            if line.startswith("#") or line.isspace():
                continue
            else:
                tokens = line.strip().split()
                s.append(float(tokens[0]))
                el.append(float(tokens[1]))
                e.append(float(tokens[2]))
    return s, el, e


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("file1", type=Path)
    parser.add_argument("file2", type=Path)

    args = parser.parse_args()

    samp1, line1, et1 = read_data(args.file1)
    samp2, line2, et2 = read_data(args.file2)

    plt.ioff()
    fig = plt.figure(constrained_layout=True)
    gs = fig.add_gridspec(3, 1)

    axtop = fig.add_subplot(gs[0, 0])
    axtop.set_xlabel("ET")
    axtop.set_ylabel("Sample")

    axbot = fig.add_subplot(gs[1, 0])
    axbot.set_xlabel("ET")
    axbot.set_ylabel("Line")

    axdiff = fig.add_subplot(gs[2, 0])
    axdiff.set_ylabel("Diff")

    axtop.plot(et1, samp1, "o", c="red")
    axtop.plot(et2, samp2, "o", c="blue", ms=2.5)

    axbot.plot(et1, line1, "o", c="red")
    axbot.plot(et2, line2, "o", c="blue", ms=2.5)

    # l1arr = np.array(line1[:-1])
    # l1arr = np.array(line1)
    # l2arr = np.array(line2)
    l2interp = np.interp(et1, et2, line2)

    # s1arr = np.array(samp1[:-1])
    # s1arr = np.array(samp1)
    # s2arr = np.array(samp2)
    s2interp = np.interp(et1, et2, samp2)

    # linediff = l1arr - l2arr
    # sampdiff = s1arr - s2arr
    linediff = line1 - l2interp
    sampdiff = samp1 - s2interp

    ldmax = np.max(np.abs(linediff))
    sdmax = np.max(np.abs(sampdiff))

    axdiff.plot(et1, linediff, c="purple", label=f"Line Diff Max: {ldmax}")
    axdiff.plot(et1, sampdiff, c="orange", label=f"Sample Diff Max: {sdmax}")
    axdiff.legend()

    plt.show()



#!/usr/bin/env python
"""This module has tests for the HiRISE HiPrecisionInit functions."""

# Copyright 2020-2021, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import unittest
from unittest.mock import patch

import hiproc.HiPrecisionInit as hpi


class TestMock(unittest.TestCase):
    def test_arg_parser(self):
        p = hpi.arg_parser()
        self.assertIsInstance(p, argparse.ArgumentParser)

    @patch("hiproc.HiPrecisionInit.pvl.load")
    @patch("hiproc.HiPrecisionInit.arg_parser")
    @patch("hiproc.HiPrecisionInit.util.set_logger")
    @patch("hiproc.HiPrecisionInit.util.main_exceptions")
    @patch("hiproc.HiPrecisionInit.print")
    @patch(
        "hiproc.HiPrecisionInit.check",
        return_value=(
            [True, True],
            [5, 5],
            1
        )
    )
    def test_main(self, m_check, m_print, m_except, m_logger, m_parse, m_load):
        with patch("hiproc.HiPrecisionInit.check", return_value=(
            [True, True],
            [5, 5],
            1
        )):
            self.assertIsNone(hpi.main())

        with patch("hiproc.HiPrecisionInit.check", return_value=(
            [False, False],
            [5, 5],
            1
        )):
            self.assertIsNone(hpi.main())


    @patch("hiproc.HiPrecisionInit.sstats.Polyfit", return_value=("_", 5, "_"))
    def test_needs_HiJACK(self, sstats):
        need, avediff = hpi.needs_HiJACK("dummy_path", 1)
        self.assertEqual(5, avediff)
        self.assertTrue(need)
        need, _ = hpi.needs_HiJACK("dummy_path", 10)
        self.assertFalse(need)

    @patch("hiproc.HiPrecisionInit.needs_HiJACK", return_value=(True, 5))
    def test_check(self, needs):
        truth_thresh = 1
        yes_HiJACK, avediffs, thresh = hpi.check(
            ["dummy1", "dummy2"],
            dict(HiPrecisionInit=dict(
                Mean_Jitter_Magnitude_Threshold=truth_thresh
            ))
        )
        self.assertEqual([True, True], yes_HiJACK)
        self.assertEqual([5, 5], avediffs)
        self.assertEqual(truth_thresh, thresh)



#!/usr/bin/env python
"""Create an ISIS cube from a HiRISE EDR .img file and record some
statistics.

These are the functionalities that the EDR_Stats pipeline performs that are
reproduced here:

* Convert a HiRISE EDR Product to an ISIS cube for subsequent pipeline
  processing using the ISIS system for much of the work.
* Gather image statistics about the observation and write them out
  to a .json file for other programs to use.
* Calculate the Signal-to-Noise Ratio (SNR).

In general, this program is the first, simple step of the HiRISE processing
chain, and prepares data for the next step: HiCal.


Data Flow
---------
Input Products:

- Any HiRISE PDS EDR, typically ending in ``.img``.

Output Products:

- A ``.cub`` file for each ``.img`` file processed.
- A ``.json`` file for each ``.img`` file processed containing summary
    information about the image.

"""

# Copyright 2004-2020, Arizona Board of Regents on behalf of the Lunar and
# Planetary Laboratory at the University of Arizona.
#   - Orignal Perl program.
#
# Copyright 2020-2021, Ross A. Beyer (rbeyer@seti.org)
#   - Elements of this Python program are are based on the original Perl
#     but the logic here is rewritten from scratch to emulate functionality.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This Python program is based on EDR_Stats version 2.18.2 (2021/04/09)
# and on the Perl EDR_Stats program ($Revision: 1.44 $
#                                    $Date: 2021/04/09 22:56:18 $
# and on the check_lut.pm program (2.4, 2020/04/28)
# by Eric Eliason, Audrie Fennema, R. King, and Richard Leis as employees of
# the University of Arizona.

import argparse
import concurrent.futures
import hashlib
import json
import logging
import math
import os
import pkg_resources
import sys
from itertools import repeat
from pathlib import Path

import pvl.new as pvl  # Need the new PVL objects, which are pickleable.
import kalasiris as isis

import hiproc.hirise as hirise
import hiproc.util as util

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__,
        parents=[util.parent_parser()],
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "-o", "--output",
        required=False,
        default=".EDR_Stats.cub",
        help="Output filename.  Optionally, if it starts with a '.' it is "
             "considered a suffix and will be swapped with the input"
             "file's suffix to determine the file to write. "
             "Default: %(default)s"
    )
    parser.add_argument(
        "--db",
        required=False,
        default=".HiCat.json",
        help="The .json file to output.  Optionally, if it "
        "starts with a '.' it is considered a suffix and "
        "will be swapped with the input file's suffix to "
        "determine the file to write. Default: %(default)s",
    )
    parser.add_argument(
        "--histmin",
        required=False,
        default=0.01,
        help="The minimum percentage above which DN are counted.  This only "
             "affects the value of STD_DN_LEVELS in the output .json file. "
             "If it is lower, more DN will be counted. Default: %(default)s"
    )
    parser.add_argument(
        "--histmax",
        required=False,
        default=99.99,
        help="The minimum percentage below which DN are counted. See --histmin "
             "for more information. Default: %(default)s"
    )
    parser.add_argument(
        "-g",
        "--gains",
        required=False,
        type=argparse.FileType('r'),
        default=pkg_resources.resource_stream(
            __name__,
            'data/EDR_Stats_gains_config.pvl'
        ),
        help="Path to the gains config PVL file.  Defaults to "
             "EDR_Stats_gains_config.pvl distributed with the library.",
    )
    parser.add_argument(
        "--max_workers",
        default=None,
        type=int,
        help="If more than one image is provided to work on, this program "
             "will engage multiprocessing to parallelize the work.  This "
             "multiprocessing will default to the number of processors on the "
             "machine.  If you want to throttle this to use less resources on "
             "your machin, indicate the number of processors you want to use."
    )
    parser.add_argument(
        "img",
        metavar="some.img-file",
        nargs="+",
        help="More than one can be listed here.",
    )
    return parser


def main():
    args = arg_parser().parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    if len(args.img) > 1:
        if not args.output.startswith("."):
            logger.critical(
                "With more than one input IMG file, the --output must start "
                f"with a period, and it does not: {args.output}"
            )
            sys.exit()

        if not args.db.startswith("."):
            logger.critical(
                "With more than one input IMG file, the --db must start with "
                f"a period, and it does not: {args.db}"
            )
            sys.exit()

    gainsinfo = pvl.load(args.gains)

    with util.main_exceptions(args.verbose):
        if len(args.img) == 1:
            # No need to fire up the multiprocessing for just one image.
            db_path = write_json(
                EDR_Stats(
                    args.img[0],
                    util.path_w_suffix(args.output, args.img[0]),
                    gainsinfo,
                    args.histmin,
                    args.histmax,
                    keep=args.keep
                ),
                args.db,
                args.img[0],
            )
            logger.info(f"Wrote {db_path}")
        else:
            with concurrent.futures.ProcessPoolExecutor(
                max_workers=args.max_workers
            ) as executor:
                for img, histats, in zip(args.img, executor.map(
                    EDR_Stats,
                    args.img,
                    map(util.path_w_suffix, repeat(args.output), args.img),
                    repeat(gainsinfo),
                    repeat(args.histmin),
                    repeat(args.histmax),
                    repeat(args.keep),
                )):
                    db_path = write_json(histats, args.db, img)
                    logger.info(f"Wrote {db_path}")
    return


def write_json(d: dict, outpath: str, template_path: os.PathLike) -> Path:
    """
    Writes out a Python dict as JSON to *outpath*.

    :param d: Python dictionary to serialize to a JSON file.
    :param outpath: If it starts with a "." assume it is a suffix that
        should be swapped with the suffix on *img* to get the output filename,
        otherwise use as an outpath.
    :param template_path: Pathlike that could have its suffix replaced.
    """
    json_path = util.path_w_suffix(outpath, template_path)

    with open(json_path, "w") as f:
        json.dump(d, f, indent=0, sort_keys=True)

    return json_path


def EDR_Stats(
    img: os.PathLike,
    out_path: os.PathLike,
    gainsinfo: dict,
    histmin=0.01,
    histmax=99.99,
    keep=False,
) -> dict:
    cid = hirise.get_ChannelID_fromfile(img)

    logger.info(f"{cid}: EDR_Stats start: {img}")
    try:
        logger.info(
            f"{cid}: The LUT for this file is: " + str(check_lut(img))
        )
    except KeyError as err:
        logger.error(
            f"{cid}: The LUT header area is either corrupted or has a gap."
        )
        raise err

    # Convert to .cub
    isis.hi2isis(img, to=out_path)

    histat_complete = isis.histat(
        out_path,
        useoffsets=True,
        leftimage=0,
        rightimage=1,
        leftcalbuffer=3,
        rightcalbuffer=1,
        leftcaldark=3,
        rightcaldark=1,
        leftbuffer=3,
        rightbuffer=1,
        leftdark=3,
        rightdark=1,
    )
    histats = parse_histat(histat_complete.stdout)

    # Get some info from the new cube:
    histats["PRODUCT_ID"] = isis.getkey_k(out_path, "Archive", "ProductId")
    histats["IMAGE_LINES"] = int(
        isis.getkey_k(out_path, "Dimensions", "Lines")
    )
    histats["LINE_SAMPLES"] = int(
        isis.getkey_k(out_path, "Dimensions", "Samples")
    )
    histats["BINNING"] = int(isis.getkey_k(out_path, "Instrument", "Summing"))

    histats["STD_DN_LEVELS"] = get_dncnt(out_path, histmin, histmax, keep=keep)
    histats["IMAGE_SIGNAL_TO_NOISE_RATIO"] = calc_snr(
        out_path, gainsinfo, histats, cid=cid
    )
    histats["GAP_PIXELS_PERCENT"] = (
        histats["GAP_PIXELS"]
        / (int(histats["IMAGE_LINES"]) * int(histats["LINE_SAMPLES"]))
    ) * 100.0

    tdi_bin_check(out_path, histats, cid=cid)
    lut_check(out_path, histats)

    logger.info(f"{cid}: EDR_Stats done: {out_path}")
    return histats


def parse_histat(pvltext: str) -> dict:
    """Parse the output of histat into a dictionary"""

    p = pvl.loads(pvltext)
    d = dict()

    # Image Area Statistics
    d["IMAGE_MEAN"] = p["IMAGE"]["Average"]
    d["IMAGE_STANDARD_DEVIATION"] = p["IMAGE"]["StandardDeviation"]
    d["IMAGE_MINIMUM"] = p["IMAGE"]["Minimum"]
    d["IMAGE_MAXIMUM"] = p["IMAGE"]["Maximum"]
    d["GAP_PIXELS"] = p["IMAGE"]["NullPixels"]
    d["LOW_SATURATED_PIXELS"] = p["IMAGE"]["LisPixels"]
    d["HIGH_SATURATED_PIXELS"] = p["IMAGE"]["HisPixels"]

    # Calibration Reverse-readout Statistics
    d["CAL_REVERSE_MEAN"] = p["CAL_REVERSE"]["Average"]
    d["CAL_REVERSE_STANDARD_DEVIATION"] = p["CAL_REVERSE"]["StandardDeviation"]
    d["CAL_REVERSE_MINIMUM"] = p["CAL_REVERSE"]["Minimum"]
    d["CAL_REVERSE_MAXIMUM"] = p["CAL_REVERSE"]["Maximum"]
    d["CAL_REVERSE_LIS"] = p["CAL_REVERSE"]["LisPixels"]

    # Calibration Mask Statistics
    d["CAL_MASK_MEAN"] = p["CAL_MASK"]["Average"]
    d["CAL_MASK_STANDARD_DEVIATION"] = p["CAL_MASK"]["StandardDeviation"]
    d["CAL_MASK_MINIMUM"] = p["CAL_MASK"]["Minimum"]
    d["CAL_MASK_MAXIMUM"] = p["CAL_MASK"]["Maximum"]

    # Calibration Ramp Statistics
    d["CAL_RAMP_MEAN"] = p["CAL_RAMP"]["Average"]
    d["CAL_RAMP_STANDARD_DEVIATION"] = p["CAL_RAMP"]["StandardDeviation"]
    d["CAL_RAMP_MINIMUM"] = p["CAL_RAMP"]["Minimum"]
    d["CAL_RAMP_MAXIMUM"] = p["CAL_RAMP"]["Maximum"]

    # Image Dark Reference Statistics
    d["IMAGE_DARK_MEAN"] = p["IMAGE_DARK"]["Average"]
    d["IMAGE_DARK_STANDARD_DEVIATION"] = p["IMAGE_DARK"]["StandardDeviation"]
    d["IMAGE_DARK_MINIMUM"] = p["IMAGE_DARK"]["Minimum"]
    d["IMAGE_DARK_MAXIMUM"] = p["IMAGE_DARK"]["Maximum"]

    # Image Buffer Area
    d["IMAGE_BUFFER_MEAN"] = p["IMAGE_BUFFER"]["Average"]
    d["IMAGE_BUFFER_STANDARD_DEVIATION"] = p["IMAGE_BUFFER"][
        "StandardDeviation"
    ]
    d["IMAGE_BUFFER_MINIMUM"] = p["IMAGE_BUFFER"]["Minimum"]
    d["IMAGE_BUFFER_MAXIMUM"] = p["IMAGE_BUFFER"]["Maximum"]
    d["IMAGE_BUFFER_LIS"] = p["IMAGE_BUFFER"]["LisPixels"]

    # Calibration Image Dark Reference
    d["CAL_DARK_MEAN"] = p["CAL_DARK"]["Average"]
    d["CAL_DARK_STANDARD_DEVIATION"] = p["CAL_DARK"]["StandardDeviation"]
    d["CAL_DARK_MINIMUM"] = p["CAL_DARK"]["Minimum"]
    d["CAL_DARK_MAXIMUM"] = p["CAL_DARK"]["Maximum"]

    # Calibration Image Buffer Area
    d["CAL_BUFFER_MEAN"] = p["CAL_BUFFER"]["Average"]
    d["CAL_BUFFER_STANDARD_DEVIATION"] = p["CAL_BUFFER"]["StandardDeviation"]
    d["CAL_BUFFER_MINIMUM"] = p["CAL_BUFFER"]["Minimum"]
    d["CAL_BUFFER_MAXIMUM"] = p["CAL_BUFFER"]["Maximum"]
    d["CAL_BUFFER_LIS"] = p["CAL_BUFFER"]["LisPixels"]

    # Calibration Dark Ramp Area
    d["CAL_DARK_RAMP_MEAN"] = p["CAL_DARK_RAMP"]["Average"]
    d["CAL_DARK_RAMP_STANDARD_DEVIATION"] = p["CAL_DARK_RAMP"][
        "StandardDeviation"
    ]
    d["CAL_DARK_RAMP_MINIMUM"] = p["CAL_DARK_RAMP"]["Minimum"]
    d["CAL_DARK_RAMP_MAXIMUM"] = p["CAL_DARK_RAMP"]["Maximum"]

    # Image Post Ramp Area
    d["IMAGE_POST_RAMP_MEAN"] = p["IMAGE_POSTRAMP"]["Average"]
    d["IMAGE_POST_RAMP_STANDARD_DEVIATION"] = p["IMAGE_POSTRAMP"][
        "StandardDeviation"
    ]
    d["IMAGE_POST_RAMP_MINIMUM"] = p["IMAGE_POSTRAMP"]["Minimum"]
    d["IMAGE_POST_RAMP_MAXIMUM"] = p["IMAGE_POSTRAMP"]["Maximum"]

    return d


def get_dncnt(cub: os.PathLike, hmin=0.01, hmax=99.99, keep=False) -> int:
    """Extract DN count from the histogram of a cub file"""
    # I'm not sure about this method.
    # The code below is the exact logic that the original Perl program has,
    # but this is just counting the number of histogram bins
    # that are within the boundaries, not the number of DN.
    # And the # of bins is automatically computed by isis.hist,
    # so could be different for each cube.
    # logger.info(get_dncnt.__doc__)

    histfile = Path(cub).with_suffix(".hist")
    if not histfile.is_file():
        isis.hist(cub, to=histfile)

    h = isis.Histogram(histfile)

    count: int = 0
    for row in h:
        if hmin <= float(row.CumulativePercent) <= hmax:
            count += 1

    if not keep:
        histfile.unlink()
    return count


def calc_snr(
    cub: os.PathLike, gainsinfo: dict, histats: dict, cid=None
) -> float:
    """Calculate the signal to noise ratio."""
    if cid is None:
        cid = hirise.get_ChannelID_fromfile(cub)
    # logger.info(f"{cid}: " + calc_snr.__doc__)

    ccdchan = f"{cid.get_ccd()}_{cid.channel}"

    # gainspvl = pvl.load(str(gainsfile))
    gain = float(gainsinfo["Gains"][ccdchan]["Bin" + str(histats["BINNING"])])

    img_mean = float(histats["IMAGE_MEAN"])
    lis_pixels = float(histats["LOW_SATURATED_PIXELS"])
    buf_mean = float(histats["IMAGE_BUFFER_MEAN"])

    snr = -9999
    r = 90
    # Note from original file about r:
    # 150 e *Changed value to 90 e- 1/31/2012 to bring closer
    # to HIPHOP value for read noise. SM

    if 0 == lis_pixels and img_mean > 0.0 and buf_mean > 0.0:
        s = (img_mean - buf_mean) * gain
        snr = s / math.sqrt(s + r * r)
        logger.info(f"{cid}: Calculation of Signal/Noise Ratio:")
        logger.info(f"{cid}: \tIMAGE_MEAN:        {img_mean}")
        logger.info(f"{cid}: \tIMAGE_BUFFER_MEAN: {buf_mean}")
        logger.info(f"{cid}: \tR (electrons/DN):  {r}")
        logger.info(f"{cid}: \tGain:              {gain}")
        logger.info(f"{cid}: Signal/Noise ratio: {snr}")

    return snr


def check_lut(img: os.PathLike):
    """Checks whether a stored look up table (LUT) matches a known LUT."""
    # Original author of this function was Robert King, in December 2006.
    # logger.info(check_lut.__doc__)

    img_pvl = pvl.load(str(img))
    lut_type = img_pvl["INSTRUMENT_SETTING_PARAMETERS"][
        "MRO:LOOKUP_TABLE_TYPE"
    ]
    if "STORED" in lut_type:
        lut = dict()
        # dec.28.06 zero-filled LUT (ie, the image was not LUT'ed)
        lut["897256b6709e1a4da9daba92b6bde39ccfccd8c1"] = None

        # dec.19.06 build
        lut["35c8318042da3c30949f3ecb9f3876de524f158f"] = 300
        lut["874f6c26fe21577bc88f8986ec5643ff43747908"] = 301
        lut["02ae8c3834160614a65190cfee95fd62a2227447"] = 302
        lut["919d8f16363d1161b1843783c713564110cef5cc"] = 303
        lut["086adb2152c5f47741ce16e128ad3533fff1fa31"] = 304
        lut["020b40e1aab312ddf9fce61d5eb7cacd33bbeaec"] = 305
        lut["b45de94586e5ab539cc1c0b72d0e0267765e3b69"] = 306
        lut["8e4cd2ba1646d117afb9af658b9f762fcb0bcc48"] = 307
        lut["6227fbb77374afe54ba5de26a5af2cfb954f7783"] = 308
        lut["31d88a70dc4a6c9cf4991eb3f7a07c3b5df624a9"] = 309
        lut["166d23b2c7e71acebfaf07798a697670bff55df9"] = 310
        lut["d0f29aae50c21e443d7d188e7937b249fd8ba447"] = 311
        lut["a44541a295899f7fe55bf038dee9d5c130190c08"] = 312
        lut["58914a9cf58f4abbf7adc60357d8daa0d8104794"] = 313
        lut["bd416e8935556cd22e6caae1e9541ea42fc32fe7"] = 314
        lut["332a862b5963d55716bc00185906ab0195372e03"] = 315
        lut["b7d5169700fa7925adfc9ab4463c4c0c64cb9c31"] = 316
        lut["de35fb5719fabf3da400de015cf2c3155bbfe1ac"] = 317
        lut["55b78c9ca2822a6f87b3c3884d95034d4abbb725"] = 318
        lut["2aaf567e760eb313eca50cff9b3ed74f035af0d5"] = 319
        lut["43a31577737b140eecc327d09f0900c182dddbe3"] = 320
        lut["79d29aa8e13dd1991f810894821ebc518bcf3ff3"] = 321
        lut["dca0da083569fc25b910611f7d43a72738923d59"] = 322
        lut["809c4a81505eced9203e4648d7905a3de9c71f7e"] = 323
        lut["349bf694a6b11ee3e47956189aebeab234be4e9e"] = 324
        lut["063a2d3ed43bb1b47e73a38fbe49091ac4183c3c"] = 325
        lut["9ea3b154b83925f65c36888d0981b84eed35baa6"] = 326
        lut["4ca66adf72eefe25bc53edc801c1c71e99cf4ab9"] = 327

        # may.16.05 build
        lut["23f9889760466065c78b070aa83097229c391ba1"] = 200
        lut["d46c82527435d5b8a3b4a329f4a80ec80ed3a6e5"] = 201
        lut["a814e4a5a679d1ca710c92b745f79c6d2a1c2f43"] = 202
        lut["b67f6fb3adb08c25955878e5f685846c3ede6194"] = 203
        lut["7c03fd37caac821554e6c9ebe593ca60200cee4b"] = 204
        lut["6c21207a2e78517a8e069f99efd8f3c549ace95d"] = 205
        lut["b21a8cb5dfee2be1eaf5e7c8ce2b7b30c7f385ef"] = 206
        lut["801e83161358bf9a4a00a76c8970454521c7aff8"] = 207
        lut["7223df1f14d33a7a57b96d2d9d39b6ae526a1f8b"] = 208
        lut["90b01c7c6055ea580293771601e6f31cc966671c"] = 209
        lut["a23833f7421ae17c066010b1150feb696b7dbb2c"] = 210
        lut["9b20c521c050c37faa959d84e6f1031f4ad42ac6"] = 211
        lut["80b6d126c6ed96d1635bb0755ed701cefb001c72"] = 212
        lut["598df42422e6167304db3d9e2613912a62f80931"] = 213
        lut["d02cda160d162001fbd515a609837320058572c6"] = 214
        lut["876a7779d8418e1a0cb9c9a4312cb65f68fd6d0e"] = 215
        lut["fbc4af8337bd8e68659ffa9bd016ce3d9aa1892b"] = 216
        lut["2948ffacdc179dba578fcf162234a61f3ef5e19c"] = 217
        lut["010344d1afc116b92a3e5a5c3884a5c3d4f1997d"] = 218
        lut["7904b0a9c14194b6209f02161be15fe6d1416926"] = 219
        lut["a57657074cb1059b4529854e2b64db750b78ec93"] = 220
        lut["e55530adfd7b4f7bb9715bff81870f32c528c559"] = 221
        lut["e8ed81aa0c81a4d54b43d8a94fcd75133585a32c"] = 222
        lut["901e6113448d3de51fff452c67d4b9f93f700a9c"] = 223
        lut["258f9f498804b3bcfb4fea679df5aa528b71021b"] = 224
        lut["71305d279ae2f2b2915fe33ceff59bd1038f6df5"] = 225
        lut["53d91396bee5b7f8c4026343e4145b1b54269a47"] = 226
        lut["07c0e1a8453a53232ccdb09d183612dd33e28ecd"] = 227

        # sep.04.04 build
        lut["3b2797fbf2588e6815f5cef7fa33ee4d3d9ef7db"] = 100
        lut["37e4bf9213d992ce974caa936211099fbf63befc"] = 101
        lut["da37672307cb95fe51b381fad59c68c537f7dff4"] = 102
        lut["23f795926899e5bf1f32f0126418be4e91ac0f94"] = 103
        lut["3f1835f9a832cd5c31ed4eb8276a964dcbf7ce12"] = 104
        lut["95f2043d5eb69c2c5d987bccf8ffd6714cbfc333"] = 105
        lut["2cd93e0418b89329ddf94daf44bfe605884d2e7f"] = 106
        lut["51817ef40212588e80155d03e19fcdc586ef7781"] = 107
        lut["8d2ec79837e6cb684a54068218f225262cc392e3"] = 108
        lut["9a9802f68a3828a0003a302b553364fd46cc1797"] = 109
        lut["f2a728884a478c0826a8c52700a6fdc53cf645e7"] = 110
        lut["247b510cfc32ec50c2029fef7a9e2dc1e906f807"] = 111
        lut["9d37b56fec19fecc73e90c183c74a17de7e29bd1"] = 112
        lut["99e4a80b7acc14d0c903f45a379cd43d14ab471e"] = 113
        lut["c9cc0c2dce9f053a1fb00fac37540c27bbe287c3"] = 114
        lut["d1f7ec6a8a8ba2a249c04b8a05571bc352fbefd9"] = 115
        lut["9544bb6b8c664413b76d8045f0291caf74d30654"] = 116
        lut["d81ed931335a16a84c5b5335d533ea54018bcc75"] = 117
        lut["e5dc08252003d7e9b4efa7ee6491387b4e0912c0"] = 118
        lut["c4bfc2c546e588e3036333d485ede76d3619d407"] = 119
        lut["ac3e23a5bd22d6300e46e7d30671bac3593858d5"] = 120
        lut["687d908ad9821d90a5efb6a9661bce6620a21630"] = 121
        lut["16df8d98efe62be7e18a1a83e29e15dcd70390d9"] = 122
        lut["802b52209dda8c0d20a0889cb8dd4d1432bc08f2"] = 123
        lut["b054f3a6bd8ddf33c7361b73aef1357fbf0dc894"] = 124
        lut["bfdb8f8a90e3f7e1d8495a878a42cfb790bb7e88"] = 125
        lut["49584c404d0eb35f96d947b7498d90f8deb9138c"] = 126
        lut["386ef62256956ef1d42ab53fe5c08abdb6faadd2"] = 127

        with open(img, mode="rb") as f:
            f.seek(32768 + 800)
            buf = f.read(16384)

        return lut[hashlib.sha1(buf).hexdigest()]

    return None


def tdi_bin_check(cube: os.PathLike, histats: dict, cid=None):
    """This function only logs warnings and returns nothing."""

    if cid is None:
        try:
            cid = histats['PRODUCT_ID']
        except KeyError:
            cid = hirise.get_ChannelID_fromfile(cube)

    # TDI and binning check
    if float(histats["IMAGE_MEAN"]) >= 8000:
        logger.warning(
            f"{cid}: "
            f"Channel mean greater than 8000 (TDI or binning too high)."
        )
    elif float(histats["IMAGE_MEAN"]) < 2500:
        tdi = isis.getkey_k(cube, "Instrument", "Tdi")
        if tdi == "32" or tdi == "64":
            logger.warning(f"{cid}: TDI too low.")
    return


def lut_check(cube: os.PathLike, histats: dict, cid=None):
    # LUT check
    lut = int(isis.getkey_k(cube, "Instrument", "LookupTableNumber"))
    orbit_number = int(isis.getkey_k(cube, "Archive", "OrbitNumber"))
    threshhold = dict()
    if lut != -9998:
        if orbit_number > 65881:
            # After orbit 65881, RED1 RED2 RED3 moved to to DN 900-1000 offset

            threshhold["RED0"] = (
                (6814, 22),
                (5341, 23),
                (3869, 24),
                (3133, 25),
                (2397, 26),
                (1200, 27),
            )
            threshhold["RED1"] = (
                (6619, 1),
                (5116, 2),
                (3614, 3),
                (2863, 4),
                (2112, 5),
                (900, 6),
            )
            threshhold["RED2"] = threshhold["RED1"]
            threshhold["RED3"] = threshhold["RED1"]
            threshhold["RED4"] = (
                (6684, 8),
                (5191, 9),
                (3699, 10),
                (2953, 11),
                (2207, 12),
                (1000, 13)
            )
            threshhold["RED5"] = (
                (6749, 15),
                (5266, 16),
                (3784, 17),
                (3043, 18),
                (2302, 19),
                (1100, 20),
            )
            threshhold["RED6"] = threshhold["RED4"]
            threshhold["RED7"] = threshhold["RED4"]
            threshhold["RED8"] = threshhold["RED0"]
            threshhold["RED9"] = threshhold["RED4"]
            threshhold["IR10"] = threshhold["RED1"]
            threshhold["IR11"] = threshhold["RED4"]
            threshhold["BG12"] = threshhold["RED1"]
            threshhold["BG13"] = threshhold["RED1"]
        elif orbit_number > 13057:
            # After orbit 13057, IR10 moved to DN 900-1000 offset
            threshhold["RED0"] = (
                (6814, 22),
                (5341, 23),
                (3869, 24),
                (3133, 25),
                (2397, 26),
                (1200, 27),
            )
            threshhold["RED1"] = (
                (6684, 8),
                (5191, 9),
                (3699, 10),
                (2953, 11),
                (2207, 12),
                (1000, 13),
            )
            threshhold["RED2"] = threshhold["RED1"]
            threshhold["RED3"] = threshhold["RED1"]
            threshhold["RED4"] = threshhold["RED1"]
            threshhold["RED5"] = (
                (6749, 15),
                (5266, 16),
                (3784, 17),
                (3043, 18),
                (2302, 19),
                (1100, 20),
            )
            threshhold["RED6"] = threshhold["RED1"]
            threshhold["RED7"] = threshhold["RED1"]
            threshhold["RED8"] = threshhold["RED0"]
            threshhold["RED9"] = threshhold["RED1"]
            threshhold["IR10"] = (
                (6619, 1),
                (5116, 2),
                (3614, 3),
                (2863, 4),
                (2112, 5),
                (900, 6),
            )
            threshhold["IR11"] = threshhold["RED1"]
            threshhold["BG12"] = threshhold["IR10"]
            threshhold["BG13"] = threshhold["IR10"]
        elif orbit_number > 11710:
            # After orbit 11710, RED6 moved to DN 1000-1100 offset
            threshhold["RED0"] = (
                (6814, 22),
                (5341, 23),
                (3869, 24),
                (3133, 25),
                (2397, 26),
                (1200, 27),
            )
            threshhold["RED1"] = (
                (6684, 8),
                (5191, 9),
                (3699, 10),
                (2953, 11),
                (2207, 12),
                (1000, 13),
            )
            threshhold["RED2"] = threshhold["RED1"]
            threshhold["RED3"] = threshhold["RED1"]
            threshhold["RED4"] = threshhold["RED1"]
            threshhold["RED5"] = (
                (6749, 15),
                (5266, 16),
                (3784, 17),
                (3043, 18),
                (2302, 19),
                (1100, 20),
            )
            threshhold["RED6"] = threshhold["RED1"]
            threshhold["RED7"] = threshhold["RED1"]
            threshhold["RED8"] = threshhold["RED0"]
            threshhold["RED9"] = threshhold["RED1"]
            threshhold["IR10"] = threshhold["RED1"]
            threshhold["IR11"] = threshhold["RED1"]
            threshhold["BG12"] = (
                (6619, 1),
                (5116, 2),
                (3614, 3),
                (2863, 4),
                (2112, 5),
                (900, 6),
            )
            threshhold["BG13"] = threshhold["BG12"]
        elif orbit_number > 2660:
            # New LUT table after orbit 2660
            threshhold["RED0"] = (
                (6814, 22),
                (5341, 23),
                (3869, 24),
                (3133, 25),
                (2397, 26),
                (1200, 27),
            )
            threshhold["RED1"] = (
                (6684, 8),
                (5191, 9),
                (3699, 10),
                (2953, 11),
                (2207, 12),
                (1000, 13),
            )
            threshhold["RED2"] = threshhold["RED1"]
            threshhold["RED3"] = threshhold["RED1"]
            threshhold["RED4"] = threshhold["RED1"]
            threshhold["RED5"] = (
                (6749, 15),
                (5266, 16),
                (3784, 17),
                (3043, 18),
                (2302, 19),
                (1100, 20),
            )
            threshhold["RED6"] = (
                (6619, 1),
                (5116, 2),
                (3614, 3),
                (2863, 4),
                (2112, 5),
                (900, 6),
            )
            threshhold["RED7"] = threshhold["RED1"]
            threshhold["RED8"] = threshhold["RED0"]
            threshhold["RED9"] = threshhold["RED1"]
            threshhold["IR10"] = threshhold["RED1"]
            threshhold["IR11"] = threshhold["RED1"]
            threshhold["BG12"] = threshhold["RED6"]
            threshhold["BG13"] = threshhold["RED6"]
        else:
            # Original LUTs prior to 2661
            threshhold["RED0"] = (
                (14057, 21),
                (11676, 22),
                (9295, 23),
                (7152, 24),
                (5248, 25),
                (3343, 26),
                (1200, 27),
            )
            threshhold["RED1"] = (
                (13857 + 1, 14),
                (11476 + 1, 15),
                (9095 + 1, 16),
                (6952 + 1, 17),
                (5048 + 1, 18),
                (3143 + 1, 19),
                (1000, 20),
            )
            threshhold["RED2"] = threshhold["RED1"]
            threshhold["RED3"] = threshhold["RED1"]
            threshhold["RED4"] = threshhold["RED1"]
            threshhold["RED5"] = threshhold["RED1"]
            threshhold["RED6"] = (
                (13657 + 1, 7),
                (11276 + 1, 8),
                (8895 + 1, 9),
                (6752 + 1, 10),
                (4848 + 1, 11),
                (2943 + 1, 12),
                (800, 13),
            )
            threshhold["RED7"] = threshhold["RED1"]
            threshhold["RED8"] = threshhold["RED0"]
            threshhold["RED9"] = threshhold["RED1"]
            threshhold["IR10"] = threshhold["RED1"]
            threshhold["IR11"] = threshhold["RED1"]
            threshhold["BG12"] = threshhold["RED6"]
            threshhold["BG13"] = threshhold["RED6"]

        ccd = hirise.ChannelID(
            isis.getkey_k(cube, "Archive", "ProductId")
        ).get_ccd()
        if cid is None:
            cid = ccd

        for (th, ex) in threshhold[ccd]:
            if float(histats["IMAGE_MEAN"]) >= th:
                lut_diff = lut - ex
                if lut_diff >= 1 or lut_diff <= -1:
                    if lut_diff > 0:
                        direction = "to the right"
                    else:
                        direction = "to the left"
                    logger.warning(
                        f"{cid}: LUT is {lut_diff} column(s) ({direction}) "
                        f"from ideal settings - image overcompressed."
                    )
                break
        else:
            logger.warning(
                f"{cid}: DN value, {histats['IMAGE_MEAN']}, lower than lowest "
                f"DN value with defined LUT for this channel."
            )
    return


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""This module has tests for the HiRISE HiCal functions."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import csv
import pkg_resources
import unittest
from pathlib import Path
from unittest.mock import patch

import pvl

import kalasiris as isis
import kalasiris.version as isisversion
import hiproc.hirise as hirise
import hiproc.EDR_Stats as edr
import hiproc.HiCal as hc

from .utils import resource_check as rc

# Hardcoding these, but I sure would like a better solution.
HiRISE_imgs = (
    "PSP_010502_2090_RED3_0.img",
    "PSP_010502_2090_RED3_1.img",
    "PSP_010502_2090_RED4_0.img",
    "PSP_010502_2090_RED4_1.img",
    "PSP_010502_2090_RED5_0.img",
    "PSP_010502_2090_RED5_1.img",
    "PSP_010502_2090_IR10_0.img",
    "PSP_010502_2090_IR10_1.img",
    "PSP_010502_2090_IR11_0.img",
    "PSP_010502_2090_IR11_1.img",
    "PSP_010502_2090_BG12_0.img",
    "PSP_010502_2090_BG12_1.img",
    "PSP_010502_2090_BG13_0.img",
    "PSP_010502_2090_BG13_1.img",
)
test_resources = Path("test-resources")
imgs = list(map(test_resources.joinpath, HiRISE_imgs))

gains = pvl.load(
    pkg_resources.resource_stream(
        "hiproc",
        "data/EDR_Stats_gains_config.pvl"
    )
)
conf_path = pkg_resources.resource_filename("hiproc", "data/HiCal.conf")
conf = pvl.load(conf_path)
# hgf_conf = Path("data") / "HiGainFx.conf"
nf_conf = pvl.load(
    pkg_resources.resource_stream("hiproc", "data/NoiseFilter.conf")
)


class TestResources(unittest.TestCase):
    """Establishes that the test image exists."""

    def test_resources(self):
        for f in imgs:
            with self.subTest(filepath=f):
                (truth, test) = rc(f)
                self.assertEqual(truth, test)


class TestBasic(unittest.TestCase):
    def test_furrow_setup(self):
        self.assertEqual(hc.furrow_setup("RED0", 4)[0], 8000)
        self.assertRaises(KeyError, hc.furrow_setup, "RED5", 1)

    def test_samp_setup(self):
        chansamp = hc.chan_samp_setup(1, 2)
        self.assertEqual(chansamp.samp[1], 511)
        self.assertEqual(chansamp.ssamp, 1)
        self.assertEqual(chansamp.nsamp, 502)

    def test_cut_size(self):
        self.assertEqual((6, 6), hc.cut_size(0, 10))
        self.assertEqual((50, 6), hc.cut_size(0, 255))
        self.assertEqual((6, 40), hc.cut_size(1, 511))

    def test_set_lines(self):
        self.assertEqual((1, 500), hc.set_lines(0, 0, 1, 500))
        self.assertEqual((2, 3998), hc.set_lines(2, 2, 2, 4000))
        self.assertRaises(ZeroDivisionError, hc.set_lines, 0, 0, 0, 0)

    def test_process_this(self):
        flags = (
            1,
            1,
            1,
            1,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
        )
        self.assertEqual(0, hc.process_this(("RED5", 0), flags))
        self.assertEqual(1, hc.process_this(("RED0", 1), flags))
        self.assertEqual(0, hc.process_this(("BG13", 1), flags))

    def test_set_flags(self):
        my_c = {
            "HiCal_Noise_Processing": (
                1,
                1,
                1,
                1,
                0,
                0,
                1,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
            ),
            "HiCal_Noise_Bin_DarkPixel_STD": (20.0, 30.0, 70.0, 100.0, 100.0),
            "HiCal_Noise_Bin_Mask_STD": (20.0, 30.0, 70.0, 100.0, 100.0),
            "HiCal_Noise_LIS_Count": 1,
            "HiCal_HPF_Cubenorm": "DIVIDE",
        }
        my_db = {
            "IMAGE_DARK_STANDARD_DEVIATION": 18.04943853369,
            "CAL_MASK_STANDARD_DEVIATION": 17.974952301437,
            "LOW_SATURATED_PIXELS": 0,
        }
        b = 1, 2, 4, 8, 16
        self.assertEqual(
            (False, False, True),
            hc.set_flags(my_c, my_db, ("RED5", 0), b.index(2)),
        )
        my_db["LOW_SATURATED_PIXELS"] = 1
        my_c["HiCal_HPF_Cubenorm"] = "SUBTRACT"
        self.assertEqual(
            (True, True, False),
            hc.set_flags(my_c, my_db, ("RED1", 1), b.index(2)),
        )

    def test_getHistVal(self):

        if tuple(isisversion.version_info()[:3]) < (4, 3, 0):
            col = "DN"
        else:
            col = "MaxExclusive"

        histogram = isis.Histogram(
            f"""Total Pixels:    2048000
Null Pixels:     0
Lis Pixels:      0

{col},Pixels,CumulativePixels,Percent,CumulativePercent
3889,1,1,4.88281e-05,4.88281e-05
3924,1,2,4.88281e-05,9.76563e-05
3960,2,4,9.76563e-05,0.000195313
3995,1,5,4.88281e-05,0.000244141
4030,4,9,0.000195313,0.000439453
6841,17215,2020810,0.840576,98.6724
6887,9753,2030563,0.476221,99.1486
7258,460,2045676,0.0224609,99.8865
7304,364,2046040,0.0177734,99.9043
7350,279,2046319,0.013623,99.9179
7759,108,2047734,0.00527344,99.987
7826,104,2047838,0.00507813,99.9921
7893,78,2047916,0.00380859,99.9959
7961,37,2047953,0.00180664,99.9977
8028,21,2047974,0.00102539,99.9987
8095,12,2047986,0.000585937,99.9993
8163,12,2047998,0.000585937,99.9999
8230,2,2048000,9.76563e-05,100"""
        )
        conf = dict(
            NoiseFilter_HighEnd_Percent=99.999,
            NoiseFilter_Hard_Tolmax=1.5,
            NoiseFilter_Hard_HighEnd_Percent=99.9,
        )
        self.assertEqual((0, 8095), hc.getHistVal(histogram, conf))

        bad_conf = dict(conf)
        bad_conf["NoiseFilter_HighEnd_Percent"] = 150
        self.assertRaises(ValueError, hc.getHistVal, histogram, bad_conf)

        histogram.dictionary["Lis Pixels"] = int(histogram["Total Pixels"]) / 2
        self.assertEqual((50, 7304), hc.getHistVal(histogram, conf))

    def test_FurrowCheck(self):
        vpnts = [1000, 1000, 1000]
        self.assertFalse(hc.FurrowCheck(vpnts, 0))

        vpnts0 = [1] + vpnts
        self.assertTrue(hc.FurrowCheck(vpnts0, 0))
        self.assertFalse(hc.FurrowCheck(vpnts0, 1))

        vpnts1 = vpnts + [1]
        self.assertTrue(hc.FurrowCheck(vpnts1, 1))
        self.assertFalse(hc.FurrowCheck(vpnts1, 0))

    def test_Cubenorm_Filter_filter_boxfilter(self):
        mylist = [4000, 4000, 0, 4000, 4000, 3800, 4000, 4000]

        truthlist = [
            3974.1882476041565,
            3974.0124192049366,
            3973.604595524187,
            3973.0384085658607,
            3972.4160735379446,
            3971.849930095303,
            3971.442163554957,
            3971.266366671061,
        ]

        self.assertEqual(
            truthlist, hc.Cubenorm_Filter_filter_boxfilter(mylist, 2, 50)
        )


class TestMock(unittest.TestCase):
    def test_get_bins_fromfiles(self):
        cubes = list()
        for p in imgs:
            cub = p.with_suffix(".PathOnly.cub")
            cubes.append(cub)
        with patch("hiproc.HiCal.isis.getkey_k", return_value="2"):
            with patch("pathlib.Path.glob", return_value=cubes):
                d = hc.get_bins_fromfiles(cubes[0])
                self.assertEqual(len(cubes) / 2, len(d))

    def test_check_destripe(self):
        bins = dict(RED3=2, RED4=2, RED5=2, IR10=4, IR11=4, BG12=4, BG13=4)
        powered = (
            "Off, Off, Off, On, On, On, On, On, On, On, Off, Off, Off, Off"
        )

        self.assertEqual(False, hc.check_destripe("dummy", 0, True, True))
        self.assertEqual(True, hc.check_destripe("dummy", 2, True, True))
        self.assertEqual(False, hc.check_destripe("dummy", 0, False, False))
        with patch("hiproc.HiCal.get_bins_fromfiles", return_value=bins):
            with patch("hiproc.HiCal.isis.getkey_k", return_value=powered):
                self.assertEqual(
                    False, hc.check_destripe("dummy", 2, None, None)
                )


class TestConf(unittest.TestCase):
    def test_conf_check(self):
        self.assertIsNone(hc.conf_check(conf))


class TestNeedCubenormStatsFile(unittest.TestCase):
    def setUp(self):
        cube = imgs[0].with_suffix(".TestHiCal_TestNeedCNstatsfile.cub")
        self.statsfile = cube.with_suffix(".stats")
        self.output = Path("dummy.tab")

        isis.hi2isis(imgs[0], to=cube)
        isis.cubenorm(cube, stats=self.statsfile)
        cube.unlink()

    def tearDown(self):
        with contextlib.suppress(FileNotFoundError):
            self.statsfile.unlink()
            self.output.unlink()
            Path("print.prt").unlink()

    def test_analyze_cubenorm_stats(self):
        self.assertEqual(
            (2901.0, 10491.599999999999),
            hc.analyze_cubenorm_stats(self.statsfile, 2)
        )

    @patch("hiproc.HiCal.isis.cubenormfile.DictWriter")
    @patch("hiproc.HiCal.NoiseFilter_cubenorm_writer")
    def test_NoiseFilter_cubenorm_edit(self, m_writer, m_DictWriter):
        conf = dict(
            NoiseFilter_Zap_Fraction=0.4, NoiseFilter_Nonvalid_Fraction=0.90
        )

        hc.NoiseFilter_cubenorm_edit(
            self.statsfile, self.output, 0, 2, conf, True
        )
        # print(m_writer.call_args_list)

    def test_Cubenorm_Filter_filter(self):
        vpnts = list()
        averages = list()
        with open(self.statsfile) as csvfile:
            reader = csv.DictReader(csvfile, dialect=isis.cubenormfile.Dialect)
            for row in reader:
                vpnts.append(int(row.pop("ValidPoints")))
                averages.append(float(row.pop("Average")))

        self.assertAlmostEqual(
            1.0036515,
            hc.Cubenorm_Filter_filter(averages, 5, 50, 0, True, vpnts, True)[
                -1
            ],
            6,
        )

    def test_Cubenorm_Filter(self):
        with patch("hiproc.HiCal.csv.DictWriter"):
            t = hc.Cubenorm_Filter(
                self.statsfile, self.output, False, 5, False, 0
            )
            self.assertAlmostEqual(15.4995326, t[0], 6)
            self.assertFalse(t[1])


class TestNeedISISCube(unittest.TestCase):
    def setUp(self):
        self.cube = imgs[0].with_suffix(".TestHiCal_TestNeedISISCube.cub")
        isis.hi2isis(imgs[0], to=self.cube)
        self.pid = hirise.get_ChannelID_fromfile(self.cube)
        self.binning = int(isis.getkey_k(self.cube, "Instrument", "Summing"))

    def tearDown(self):
        with contextlib.suppress(FileNotFoundError):
            self.cube.unlink()
            Path("print.prt").unlink()

    def test_furrow_nulling(self):
        ccdchan = (self.pid.get_ccd(), int(self.pid.channel))
        outcube = Path("test_furrow_nulling-out.cub")
        self.assertFalse(
            hc.furrow_nulling(self.cube, outcube, self.binning, ccdchan, False)
        )
        outcube.unlink()

    def test_mask(self):
        outcube = Path("test_mask-out.cub")
        self.assertIsNone(
            hc.mask(self.cube, outcube, 1200, 16383, self.binning, False)
        )
        outcube.unlink()

    def test_run_hical(self):
        myconf = dict(HiCal=None, NoiseFilter=None)
        myconf["HiCal"] = dict(
            HiCal_Normalization_Minimum="0.0",
            HiCal_Normalization_Maximum="1.5",
            HiCal_ISIS_Conf="hical.pipelines.conf",
            HiCal_ISIS_Conf_Noise="hical.noise.pipelines.conf",
        )
        myconf["NoiseFilter"] = dict(
            NoiseFilter_Raw_Min="1200", NoiseFilter_Raw_Max="16383"
        )
        outcube = Path("test_run_hical-out.cub")

        self.assertEqual(
            "Standard",
            hc.run_hical(
                self.cube,
                outcube,
                myconf,
                conf_path,
                3,
                3,
                self.binning,
                True,
                keep=False,
            ),
        )
        outcube.unlink()

    # def test_HiGainFx(self):
    #     outcube = Path('test_run_HiGainFx-out.cub')
    #     self.assertIsNone(hc.HiGainFx(self.cube, outcube, Path('data'),
    #                                   '0001', keep=True))
    #     # outcube.unlink()

    def test_highlow_destripe(self):
        myconf = dict(
            NoiseFilter_LPF_Line=501,
            NoiseFilter_LPF_Samp=9,
            NoiseFilter_LPF_Minper=5,
            NoiseFilter_HPF_Line=501,
            NoiseFilter_HPF_Samp=1,
            NoiseFilter_HPF_Minper=5,
        )
        outcube = Path("test_highlow_destripe-out.cub")
        self.assertIsNone(
            hc.highlow_destripe(
                self.cube,
                self.cube,
                outcube,
                myconf,
                isisnorm="",
                lnull=True,
                lhrs=True,
                lhis=True,
                llrs=True,
                llis=True,
                keep=False,
            )
        )
        outcube.unlink()

    def test_NoiseFilter_noisefilter(self):
        myconf = dict(
            NoiseFilter_Minimum_Value=0,
            NoiseFilter_Noise_Samp=7,
            NoiseFilter_Noise_Line=7,
        )
        outcube = Path("test_NoiseFilter_noisefilter-out.cub")
        self.assertIsNone(
            hc.NoiseFilter_noisefilter(
                self.cube,
                outcube,
                flattol=1,
                conf=myconf,
                maxval=4000,
                tolmin=3.5,
                tolmax=3.5,
            )
        )
        outcube.unlink()

    def test_NoiseFilter(self):
        myconf = dict(
            NoiseFilter_HighEnd_Percent=99.999,
            NoiseFilter_Hard_Tolmax=1.5,
            NoiseFilter_Hard_Tolmin=1.5,
            NoiseFilter_Hard_HighEnd_Percent=99.9,
            NoiseFilter_Zap_Fraction=0.4,
            NoiseFilter_Nonvalid_Fraction=0.90,
            NoiseFilter_LPF_Line=501,
            NoiseFilter_LPF_Samp=9,
            NoiseFilter_LPF_Minper=5,
            NoiseFilter_HPF_Line=501,
            NoiseFilter_HPF_Samp=1,
            NoiseFilter_HPF_Minper=5,
            NoiseFilter_Tolmin=3.5,
            NoiseFilter_Tolmax=3.5,
            NoiseFilter_Hard_Filtering=5,
            NoiseFilter_Flattol=1,
            NoiseFilter_Minimum_Value=0,
            NoiseFilter_Noise_Samp=7,
            NoiseFilter_Noise_Line=7,
            NoiseFilter_LPFZ_Line=5,
            NoiseFilter_LPFZ_Samp=5,
        )
        outcube = Path("test_NoiseFilter-out.cub")
        self.assertIsNone(
            hc.NoiseFilter(
                self.cube,
                outcube,
                myconf,
                minimum=None,
                maximum=None,
                zapc=False,
                keep=False,
            )
        )
        outcube.unlink()

    def test_Hidestripe(self):
        to_del = isis.PathSet()
        calcube = to_del.add(Path("test_Hidestripe-out.hical.cub"))
        isis.hical(self.cube, to=calcube)
        to_del.add(Path(str(self.pid)).with_suffix(".hical.log"))
        outcube = to_del.add(Path("test_Hidestripe-out.cub"))
        samps = int(isis.getkey_k(calcube, "Dimensions", "Samples"))

        self.assertRaises(
            KeyError,
            hc.Hidestripe,
            self.cube,
            outcube,
            self.binning,
            minimum=0.0,
            maximum=1.5,
            hidcorr="ADD",
            line_samples=samps,
            keep=False,
        )

        self.assertAlmostEqual(
            0.000101402295171637,
            hc.Hidestripe(
                calcube,
                outcube,
                self.binning,
                minimum=0.0,
                maximum=1.5,
                hidcorr="ADD",
                line_samples=samps,
                keep=False,
            ),
        )
        to_del.unlink()


class TestHiCal(unittest.TestCase):
    def setUp(self):
        self.cube = imgs[0].with_suffix(".TestHiCal.cub")
        self.pid = hirise.get_ChannelID_fromfile(self.cube)
        self.db = edr.EDR_Stats(imgs[0], self.cube, gains)
        self.binning = int(isis.getkey_k(self.cube, "Instrument", "Summing"))
        self.conf = conf
        # self.conf['HiGainFx'] = pvl.load(str(hgf_conf))['HiGainFx']
        self.conf["NoiseFilter"] = nf_conf["NoiseFilter"]

    def tearDown(self):
        with contextlib.suppress(FileNotFoundError):
            self.cube.unlink()
            Path("print.prt").unlink()

    def test_HiCal(self):
        outcube = Path("test_HiCal-out.cub")
        # ccdchan = (self.pid.get_ccd(), self.pid.channel)
        db = hc.HiCal(
            self.cube,
            outcube,
            self.db,
            self.conf,
            conf_path,
            bin2=False,
            bin4=False,
            keep=False,
        )
        # print(db)
        self.assertAlmostEqual(
            db["HIGH_PASS_FILTER_CORRECTION_STANDARD_DEVIATION"],
            0.00204738,
        )
        self.assertIsNone(db["DESTRIPED_DIFFERENCE_STANDARD_DEVIATION"])
        self.assertFalse(db["zapped"])
        self.assertEqual(db["hical_status"], "Standard")
        outcube.unlink()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `util` module."""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The LICENSE file is at the top level of this library.

import argparse
import logging
import unittest
from pathlib import Path

import hiproc.util as util

from .utils import resource_check as rc

# Hardcoding these, but I sure would like a better solution.
HiRISE_img = Path("test-resources") / "PSP_010502_2090_RED5_0.img"
img = HiRISE_img


class TestResources(unittest.TestCase):
    """Establishes that the test image exists."""

    def test_resources(self):
        (truth, test) = rc(img)
        self.assertEqual(truth, test)


class TestUtil(unittest.TestCase):
    def test_parent_parser(self):
        self.assertIsInstance(util.parent_parser(), argparse.ArgumentParser)

    def test_logging(self):
        util.set_logger(loglvl="WARNING")
        logger = logging.getLogger()
        self.assertEqual(30, logger.getEffectiveLevel())

    def test_path_w_suffix(self):
        self.assertEqual("bar.foo", str(util.path_w_suffix(".foo", "bar.cub")))
        self.assertEqual(
            "foo.foo", str(util.path_w_suffix("foo.foo", "bar.cub"))
        )

    def test_pid_path_w_suffix(self):
        self.assertRaises(
            FileNotFoundError, util.pid_path_w_suffix, "foo.foo", "dummy"
        )
        self.assertRaises(
            FileNotFoundError,
            util.pid_path_w_suffix,
            ".foo",
            "PSP_010502_2090_RED5_0.dummy",
        )
        self.assertEqual(
            img, util.pid_path_w_suffix(".img", img.with_suffix(".foo.foo"))
        )

    def test_get_path(self):
        self.assertRaises(ValueError, util.get_path, "dummy")
        self.assertRaises(TypeError, util.get_path, "dummy", 42)
        self.assertRaises(
            NotADirectoryError, util.get_path, "dummy", "not_a_dir"
        )
        self.assertRaises(FileNotFoundError, util.get_path, "dummy", "hiproc")
        self.assertRaises(
            FileNotFoundError, util.get_path, "dummy", Path("hiproc")
        )
        self.assertRaises(
            FileNotFoundError, util.get_path, "dummy", ["hiproc", "tests"]
        )
        self.assertEqual(
            Path("hiproc/util.py"), util.get_path("util.py", "hiproc")
        )

    def test_conf_check_strings(self):
        self.assertIsNone(util.conf_check_strings("foo", ("YES", "NO"), "YES"))
        self.assertRaises(
            AssertionError,
            util.conf_check_strings,
            "foo",
            ("YES", "NO"),
            "MAYBE",
        )

    def test_conf_check_count(self):
        self.assertIsNone(
            util.conf_check_count("foo", 3, "what", ["one", "two", "three"])
        )
        self.assertRaises(
            AssertionError,
            util.conf_check_count,
            "foo",
            3,
            "what",
            ["one, two"],
        )

    def test_conf_check_bounds(self):
        self.assertIsNone(util.conf_check_bounds("foo", (0.1, 1), "0.5"))
        self.assertRaises(
            AssertionError, util.conf_check_bounds, "foo", (0.1, 1), "1.5"
        )

    def test_pause_slicer(self):
        self.assertEqual(slice(4, 7), util.pause_slicer(5, 3))
        self.assertEqual(slice(2, 5), util.pause_slicer(5, -3))



#!/usr/bin/env python
"""This module has tests for the HiRISE HiStitch functions."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import pkg_resources
import unittest
from pathlib import Path
from unittest.mock import call
from unittest.mock import mock_open
from unittest.mock import patch

import pvl

import hiproc.hirise as hirise
import hiproc.HiStitch as hs

conf = pvl.load(
    pkg_resources.resource_stream(
        "hiproc",
        "data/HiStitch.conf"
    )
)


class TestBasic(unittest.TestCase):
    def test_get_chids(self):
        cid_list = (
            "PSP_010502_2090_RED4_0",
            "PSP_010502_2090_RED4_1",
            "PSP_010502_2090_RED5_0",
        )
        cube_list = list()
        for p in cid_list:
            cube_list.append(p + ".EDR_Stats.HiCal.cub")
        self.assertEqual(
            cid_list[0:2], tuple(map(str, hs.get_chids(cube_list[0:2])))
        )
        self.assertEqual(cid_list[-1], str(hs.get_chids((cube_list[-1],))[0]))
        self.assertRaises(ValueError, hs.get_chids, cube_list[1:3])
        with contextlib.suppress(FileNotFoundError):
            Path("print.prt").unlink()

    def test_set_outpath(self):
        d = Path("dummy-directory")
        output = Path("foo")
        chid = hirise.CCDID("PSP_010502_2090_RED5_0")
        self.assertEqual(output, hs.set_outpath(output, chid, d))
        suffix = ".foo"
        with_suffix = d / Path("PSP_010502_2090_RED5.foo")
        self.assertEqual(with_suffix, hs.set_outpath(suffix, chid, d))


class TestConf(unittest.TestCase):
    def test_conf_check(self):
        self.assertIsNone(hs.conf_check(conf))


class TestMock(unittest.TestCase):
    def test_sort_input_cubes(self):
        cub0 = "PSP_010502_2090_RED4_0"
        cub1 = "PSP_010502_2090_RED4_1"
        with patch(
            "hiproc.HiStitch.isis.getkey_k", side_effect=[cub0[-1], cub1[-1]]
        ):
            self.assertEqual((cub0, cub1), hs.sort_input_cubes(cub0, cub1))

        with patch(
            "hiproc.HiStitch.isis.getkey_k", side_effect=[cub1[-1], cub0[-1]]
        ):
            self.assertEqual((cub0, cub1), hs.sort_input_cubes(cub1, cub0))

        with patch("hiproc.HiStitch.isis.getkey_k", return_value=0):
            self.assertRaises(RuntimeError, hs.sort_input_cubes, cub0, cub0)

        with patch("hiproc.HiStitch.isis.getkey_k", side_effect=[0, 2]):
            self.assertRaises(RuntimeError, hs.sort_input_cubes, cub0, cub1)

    def test_sort_databases(self):
        chids = (
            hirise.ChannelID("PSP_010502_2090_RED4_0"),
            hirise.ChannelID("PSP_010502_2090_RED4_1"),
        )
        dbs = ({"PRODUCT_ID": str(chids[0])}, {"PRODUCT_ID": str(chids[1])})
        _ = "dummy_path"
        with patch("hiproc.HiStitch.open", mock_open(read_data="dummy")):
            with patch("hiproc.HiStitch.json.load", side_effect=dbs):
                self.assertRaises(
                    IndexError, hs.sort_databases, (_, _, _), chids
                )
            # with patch('hiproc.HiStitch.json.load', side_effect=dbs):
            self.assertRaises(
                LookupError, hs.sort_databases, dbs, (chids[0], chids[0])
            )
            # with patch('hiproc.HiStitch.json.load', side_effect=dbs):
            self.assertEqual(dbs, hs.sort_databases(dbs, chids))
            # with patch('hiproc.HiStitch.json.load', side_effect=reversed(dbs)):
            self.assertEqual(
                dbs, hs.sort_databases(list(reversed(dbs)), chids)
            )

    @patch("hiproc.HiStitch.isis.fx")
    @patch("hiproc.HiStitch.isis.mask")
    @patch("hiproc.HiStitch.isis.lowpass")
    @patch("hiproc.HiStitch.isis.highpass")
    @patch("hiproc.HiStitch.isis.algebra")
    @patch("hiproc.HiStitch.isis.handmos")
    @patch("shutil.copyfile")
    def test_HiFurrow_Fix(
        self,
        mock_copyfile,
        mock_handmos,
        mock_algebra,
        mock_highpass,
        mock_lowpass,
        mock_mask,
        mock_fx,
    ):
        with patch("hiproc.HiStitch.isis.getkey_k", return_value=1):
            self.assertRaises(
                ValueError, hs.HiFurrow_Fix, "dum_in", "dum_out", 0
            )
        with patch("hiproc.HiStitch.isis.getkey_k", side_effect=[2, 1, 1]):
            self.assertRaises(
                ValueError, hs.HiFurrow_Fix, "dum_in", "dum_out", 0
            )
        with patch(
            "hiproc.HiStitch.isis.getkey_k", side_effect=["2", "1024", "1024"]
        ):
            in_cube = "dummy_in.cub"
            out_cube = "dummy_out.cub"
            hs.HiFurrow_Fix(in_cube, out_cube, 1000, keep=True)

            mock_fx.assert_called_once()
            eqn = r"\(1*(sample<512)+ 1*(sample>513) + 0)"
            fx_path = mock_fx.call_args[1]["to"]
            mock_fx.assert_called_once_with(
                equation=eqn,
                lines=1024,
                mode="OUTPUTONLY",
                samples=1024,
                to=fx_path,
            )

            mask1_path = mock_mask.call_args_list[0][1]["to"]
            mask2_path = mock_mask.call_args_list[1][1]["to"]
            mock_calls = [
                call(
                    Path(in_cube),
                    mask=fx_path,
                    max_=1,
                    min_=1,
                    preserve="INSIDE",
                    spixels="NULL",
                    to=mask1_path,
                ),
                call(
                    Path(in_cube),
                    mask=fx_path,
                    max_=0,
                    min_=0,
                    preserve="INSIDE",
                    spixels="NULL",
                    to=mask2_path,
                ),
            ]
            self.assertEqual(mock_calls, mock_mask.call_args_list)

            lpf_path = mock_lowpass.call_args[1]["to"]
            mock_lowpass.assert_called_once_with(
                mask1_path,
                his=False,
                hrs=False,
                line=41,
                lis=False,
                null=True,
                sample=5,
                to=lpf_path,
            )

            hpf_path = mock_highpass.call_args[1]["to"]
            mock_highpass.assert_called_once_with(
                mask2_path, line=41, sample=1, to=hpf_path
            )

            alg_path = mock_algebra.call_args[1]["to"]
            mock_algebra.assert_called_once_with(
                A=1.0,
                B=1.0,
                from2=hpf_path,
                from_=lpf_path,
                operator="ADD",
                to=alg_path,
            )

            mock_handmos.assert_called_once_with(
                alg_path,
                create="NO",
                inband=1,
                inline=1,
                insample=1,
                mosaic=out_cube,
                outband=1,
                outline=1,
                outsample=1,
            )


class TestHiStitch(unittest.TestCase):
    def setUp(self):
        self.my_c = {
            "HiStitch_Furrow_Normalization": 4000,
            "HiStitch_Balance": True,
            "HiStitch_Equalize": False,
            "HiStitch_Balance_Processing": (
                1,
                1,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                1,
                1,
                0,
                0,
                0,
            ),
            "HiStitch_LIS_Pixels": 10000,
            "HiStitch_Balance_Bin_Mask_STD": (
                200.0,
                200.0,
                100.0,
                100.0,
                100.0,
            ),
            "HiStitch_Balance_Bin_DarkPixel_STD": (
                70.0,
                70.0,
                70.0,
                70.0,
                70.0,
            ),
            "HiStitch_Gap_Percent": 10.0,
            "HiStitch_Bin01_Area": (7, 11),
            "HiStitch_Bin02_Area": (7, 7),
            "HiStitch_Bin04_Area": (7, 5),
            "HiStitch_Control_Channel": (
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
            ),
            "HiStitch_Equalize_Width": 1501,
            "HiStitch_Equalize_Correction": "MULTIPLY",
            "HiStitch_Clean_Files": "KEEP",
            "HiStitch_Normalization_Minimum": 0,
            "HiStitch_Normalization_Maximum": 16000,
            "HiStitch_Minimum_Percent": 0
        }
        self.my_db0 = {
            "IMAGE_MEAN": 6491.34508964,
            "LOW_SATURATED_PIXELS": 0,
            "CAL_MASK_STANDARD_DEVIATION": 17.97,
            "IMAGE_DARK_STANDARD_DEVIATION": 18.05,
            "GAP_PIXELS_PERCENT": 0,
            "BINNING": 2,
            "zapped": False,
        }
        self.my_db1 = {
            "IMAGE_MEAN": 9.34508964,
            "LOW_SATURATED_PIXELS": 0,
            "CAL_MASK_STANDARD_DEVIATION": 17.97,
            "IMAGE_DARK_STANDARD_DEVIATION": 18.05,
            "GAP_PIXELS_PERCENT": 0,
            "BINNING": 2,
            "zapped": False,
        }
        self.my_dbs = [self.my_db0, self.my_db1]

    def test_set_flags(self):
        b = 1, 2, 4, 8, 16
        self.assertEqual(
            (True, True, False),
            hs.set_flags(self.my_c, self.my_dbs, 5, b.index(2), 6491.35),
        )
        self.my_c["HiStitch_Equalize"] = True
        self.assertEqual(
            (True, False, True),
            hs.set_flags(self.my_c, self.my_dbs, 5, b.index(2), 6491.35),
        )
        self.my_dbs[0]["zapped"] = True
        self.assertEqual(
            (False, False, True),
            hs.set_flags(self.my_c, self.my_dbs, 5, b.index(2), 6491.35),
        )

    def test_HiStitchStep(self):
        cubes = ("dummy1.in", "dummy2.in")
        out_c = "dummy.out"

        with patch("hiproc.HiStitch.isis.histitch") as mock:
            hs.HiStitchStep(cubes, out_c, 2, 5, self.my_c, False, False)
            mock.assert_called_with(
                from1=cubes[0], from2=cubes[1], to=out_c, balance=False
            )

            hs.HiStitchStep([cubes[0], ], out_c, 2, 5, self.my_c, False, False)
            mock.assert_called_with(from1=cubes[0], to=out_c, balance=False)

            hs.HiStitchStep(cubes, out_c, 2, 5, self.my_c, True, False)
            mock.assert_called_with(
                from1=cubes[0],
                from2=cubes[1],
                to=out_c,
                balance="TRUE",
                channel=0,
                seamsize=7,
                skip=7,
            )

            hs.HiStitchStep(cubes, out_c, 4, 0, self.my_c, False, True)
            mock.assert_called_with(
                from1=cubes[0],
                from2=cubes[1],
                to=out_c,
                balance="EQUALIZE",
                channel=1,
                seamsize=5,
                skip=7,
                operator="MULTIPLY",
                width=1501,
            )

    @unittest.skip("Need more work to make a good test here.")
    @patch(
        "hiproc.hirise.get_ChannelID_fromfile",
        return_value="PSP_010502_2090_RED4_0"
    )
    @patch("hiproc.HiStitch.Path.rename")
    @patch("hiproc.HiStitch.isis.fx")
    @patch("hiproc.HiStitch.isis.mask")
    @patch("hiproc.HiStitch.isis.lowpass")
    @patch("hiproc.HiStitch.isis.highpass")
    @patch("hiproc.HiStitch.isis.algebra")
    @patch("hiproc.HiStitch.isis.handmos")
    @patch("hiproc.HiStitch.isis.histitch")
    @patch("shutil.copyfile")
    def test_HiStitch(
        self,
        mock_histitch,
        mock_copyfile,
        mock_handmos,
        mock_algebra,
        mock_highpass,
        mock_lowpass,
        mock_mask,
        mock_fx,
        mock_rename,
        mock_cid_fromfile
    ):
        cubes = ("dummy1.in", "dummy2.in")
        out_c = Path("dummy.out")

        def getkey(cube, group, key):
            values = {
                "TruthChannel": 0,  # need to look at actual output
                "BalanceRatio": 3,
                "Summing": 2,
                "Lines": 1024,
                "Samples": 1024,
                "PRODUCT_ID": "PSP_010502_2090_RED4_0"
            }
            if key == "ChannelNumber":
                return cube.stem[-1]
            else:
                return values[key]

        db0 = self.my_db0
        db0["PRODUCT_ID"] = "PSP_010502_2090_RED4_0"
        db1 = self.my_db1
        db1["PRODUCT_ID"] = "PSP_010502_2090_RED4_1"

        with patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey):
            self.assertEqual(
                (0, 3),
                hs.HiStitch(
                    Path("dummy0.in"),
                    Path("dummy1.in"),
                    db0,
                    db1,
                    out_c,
                    {"HiStitch": self.my_c},
                    keep=True
                ),
            )

        with patch("hiproc.HiStitch.isis.getkey_k", side_effect=getkey):
            self.assertEqual(
                (None, None),
                hs.HiStitch(
                    [cubes, ], out_c, self.my_c, self.my_dbs, 5, keep=True
                ),
            )



#!/usr/bin/env python
"""This module has tests for the HiRISE HiccdStitch functions."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import contextlib
import unittest
from pathlib import Path
from unittest.mock import patch

import hiproc.HiColorInit as hci


def getkey(cube, group, key):
    values = {
        "ProductId": None,
        "Summing": 2,
        "Lines": 1024,
        "Samples": 1024,
        "TDI": 64,
    }
    return values[key]


class TestHiColorCube(unittest.TestCase):
    @patch("hiproc.HiColorInit.isis.getkey_k", side_effect=getkey)
    def test_init(self, mock_getkey):
        c = hci.HiColorCube("dummy/PSP_010502_2090_RED5_0")
        self.assertTrue(c.tdi, 64)


class TestMock(unittest.TestCase):
    @patch("hiproc.HiColorInit.isis.getkey_k", side_effect=getkey)
    def test_separate_ccds(self, mock_getkey):
        c04 = hci.HiColorCube("dummy/PSP_010502_2090_RED4")
        c05 = hci.HiColorCube("dummy/PSP_010502_2090_RED5")
        c10 = hci.HiColorCube("dummy/PSP_010502_2090_IR10")
        c11 = hci.HiColorCube("dummy/PSP_010502_2090_IR11")
        c12 = hci.HiColorCube("dummy/PSP_010502_2090_BG12")
        c13 = hci.HiColorCube("dummy/PSP_010502_2090_BG13")

        (red4, red5, ir10, ir11, bg12, bg13) = hci.separate_ccds(
            [c04, c05, c10, c11, c12, c13]
        )
        self.assertEqual(c04, red4)
        self.assertEqual(c05, red5)
        self.assertEqual(c10, ir10)
        self.assertEqual(c11, ir11)
        self.assertEqual(c12, bg12)
        self.assertEqual(c13, bg13)

    @patch("hiproc.HiColorInit.isis.reduce")
    @patch("hiproc.HiColorInit.isis.enlarge")
    @patch("hiproc.HiColorInit.isis.handmos")
    @patch("hiproc.HiColorInit.isis.editlab")
    @patch("hiproc.HiColorInit.Path.unlink")
    @patch("hiproc.HiColorInit.isis.getkey_k", side_effect=getkey)
    def test_HiColorInit_reduce1(
        self, m_getkey, m_unlink, m_editlab, m_handmos, m_enlarge, m_reduce
    ):
        red = Path("dummy/PSP_010502_2090_RED4")
        ir = Path("dummy/PSP_010502_2090_IR10")
        bg = Path("dummy/PSP_010502_2090_BG12")
        suffix = ".dummy.cub"
        hci.HiColorInit([red, ir, bg], suffix, keep=True)
        self.assertEqual(2, m_reduce.call_count)
        self.assertAlmostEqual(m_reduce.call_args[1]["sscale"], 1.0006)
        m_enlarge.assert_not_called()
        self.assertEqual(2, m_editlab.call_count)
        self.assertTrue(str(m_editlab.call_args[0][0]).endswith(suffix))
        self.assertAlmostEqual(m_handmos.call_args[1]["nlines"], 1024)
        m_unlink.assert_not_called()

    @patch("hiproc.HiColorInit.isis.reduce")
    @patch("hiproc.HiColorInit.isis.enlarge")
    @patch("hiproc.HiColorInit.isis.handmos")
    @patch("hiproc.HiColorInit.isis.editlab")
    @patch("hiproc.HiColorInit.Path.unlink")
    @patch("hiproc.HiColorInit.isis.getkey_k", side_effect=getkey)
    def test_HiColorInit_reduce2(
        self, m_getkey, m_unlink, m_editlab, m_handmos, m_enlarge, m_reduce
    ):
        red = Path("dummy/PSP_010502_2090_RED4")
        ir = Path("dummy/PSP_010502_2090_IR10")
        s = ".dummy.cub"
        hci.HiColorInit([red, ir], s)
        m_reduce.assert_called_once()
        self.assertAlmostEqual(1.0006, m_reduce.call_args[1]["sscale"])
        m_enlarge.assert_not_called()
        m_unlink.assert_called_once()
